#!/usr/bin/env python3
# coding: utf-8

"""
Unit tests for the persistent_functions submodule in corepers module
"""
import os
import logging

import PanACoTA.corepers_module.persistent_functions as persf
import test.test_unit.utilities_for_tests as tutils


PERS_PATH = os.path.join("test", "data", "persgenome")
EXP_PATH = os.path.join(PERS_PATH, "exp_files")
FAMS_BY_STRAIN = \
    {'1': {'GEN4.1111.00001': ['GEN4.1111.00001.b0001_00001'],
           'GENO.0817.00001': ['GENO.0817.00001.b0001_00002'],
           'GENO.1216.00002': ['GENO.1216.00002.b0001_00001', 'GENO.1216.00002.i0001_00002'],
           'GENO.1216.00003': ['GENO.1216.00003.i0001_00003']
           },
     '10': {'GEN4.1111.00001': ['GEN4.1111.00001.i0001_00004'],
            'GENO.0817.00001': ['GENO.0817.00001.i0002_00004'],
            'GENO.1216.00002': ['GENO.1216.00002.i0001_00005']
            },
     '11': {'GEN4.1111.00001': ['GEN4.1111.00001.i0001_00005'],
            'GENO.0817.00001': ['GENO.0817.00001.i0002_00005'],
            'GENO.1216.00002': ['GENO.1216.00002.i0001_00006']
            },
     '12': {'GEN4.1111.00001': ['GEN4.1111.00001.i0001_00008'],
            'GENO.0817.00001': ['GENO.0817.00001.i0002_00010'],
            'GENO.1216.00002': ['GENO.1216.00002.b0002_00009'],
            'GENO.1216.00003': ['GENO.1216.00003.i0001_00004', 'GENO.1216.00003.i0001_01000']
            },
     '13': {'GENO.1216.00002': ['GENO.1216.00002.b0003_00011'],
            'GENO.0817.00001': ['GENO.0817.00001.i0002_00010']},
     '14': {'GENO.1216.00002': ['GENO.1216.00002.b0003_00012']},
     '2': {'GEN4.1111.00001': ['GEN4.1111.00001.i0001_00003']},
     '3': {'GEN4.1111.00001': ['GEN4.1111.00001.b0001_00009'],
           'GENO.0817.00001': ['GENO.0817.00001.b0002_00011'],
           'GENO.1216.00002': ['GENO.1216.00002.b0002_00010'],
           'GENO.1216.00003': ['GENO.1216.00003.i0001_01010']
           },
     '4': {'GENO.0817.00001': ['GENO.0817.00001.b0001_00001']},
     '5': {'GEN4.1111.00001': ['GEN4.1111.00001.i0001_00002'],
           'GENO.0817.00001': ['GENO.0817.00001.b0002_00003'],
           'GENO.1216.00002': ['GENO.1216.00002.i0001_00003'],
           'GENO.1216.00003': ['GENO.1216.00003.i0080_00010']
           },
     '6': {'GEN4.1111.00001': ['GEN4.1111.00001.i0001_00006'],
           'GENO.0817.00001': ['GENO.0817.00001.i0002_00006', 'GENO.0817.00001.i0002_00007'],
           'GENO.1216.00002': ['GENO.1216.00002.i0001_00007']
           },
     '7': {'GENO.0817.00001': ['GENO.0817.00001.i0002_00008']},
     '8': {'GEN4.1111.00001': ['GEN4.1111.00001.i0001_00007'],
           'GENO.0817.00001': ['GENO.0817.00001.i0002_00009'],
           'GENO.1216.00002': ['GENO.1216.00002.b0001_00008']
           },
     '9': {'GENO.1216.00002': ['GENO.1216.00002.i0001_00004']}
     }

FAMILIES = {'1': ['GEN4.1111.00001.b0001_00001', 'GENO.0817.00001.b0001_00002',
                  'GENO.1216.00002.b0001_00001', 'GENO.1216.00002.i0001_00002',
                  'GENO.1216.00003.i0001_00003'],
            '10': ['GEN4.1111.00001.i0001_00004', 'GENO.0817.00001.i0002_00004',
                   'GENO.1216.00002.i0001_00005'],
            '11': ['GEN4.1111.00001.i0001_00005', 'GENO.0817.00001.i0002_00005',
                   'GENO.1216.00002.i0001_00006'],
            '12': ['GEN4.1111.00001.i0001_00008', 'GENO.0817.00001.i0002_00010',
                   'GENO.1216.00002.b0002_00009', 'GENO.1216.00003.i0001_00004',
                   'GENO.1216.00003.i0001_01000'],
            '13': ['GENO.1216.00002.b0003_00011', 'GENO.0817.00001.i0002_00010'],
            '14': ['GENO.1216.00002.b0003_00012'],
            '2': ['GEN4.1111.00001.i0001_00003'],
            '3': ['GEN4.1111.00001.b0001_00009', 'GENO.0817.00001.b0002_00011',
                  'GENO.1216.00002.b0002_00010', 'GENO.1216.00003.i0001_01010'],
            '4': ['GENO.0817.00001.b0001_00001'],
            '5': ['GEN4.1111.00001.i0001_00002', 'GENO.0817.00001.b0002_00003',
                  'GENO.1216.00002.i0001_00003', 'GENO.1216.00003.i0080_00010'],
            '6': ['GEN4.1111.00001.i0001_00006', 'GENO.0817.00001.i0002_00006',
                  'GENO.0817.00001.i0002_00007', 'GENO.1216.00002.i0001_00007'],
            '7': ['GENO.0817.00001.i0002_00008'],
            '8': ['GEN4.1111.00001.i0001_00007', 'GENO.0817.00001.i0002_00009',
                  'GENO.1216.00002.b0001_00008'],
            '9': ['GENO.1216.00002.i0001_00004']
            }


def test_uniq_mems():
    """
    Test that it returns True when there is only 1 member of each genome in the given family
    """
    family = {"strain1": ["member"],
              "strain2": ["member2"],
              "strain3": ["member3"]}
    assert persf.uniq_members(family)


def test_nonuniq_mems():
    """
    Test that it returns False when there is more than 1 member in a genome
    """
    family = {"strain1": ["member", "member-bis"],
              "strain2": ["member2"],
              "strain3": ["member3"]}
    assert not persf.uniq_members(family)


def test_zero_mem(caplog):
    """
    Test that when there are no member in 1 family, and 1 in all others, it
    returns True but with a warning message.
    """
    caplog.set_level(logging.DEBUG)
    family = {"strain1": ["member"],
              "strain2": ["member2"],
              "strain3": ["member3"],
              "strain4": []}
    assert persf.uniq_members(family)
    assert "Problem, no members for strain4!" in caplog.text


def test_zero_mem_notuniq(caplog):
    """
    Test that when there are no member in 1 family, and several in 1 other family, it
    returns False + a warning message.
    """
    caplog.set_level(logging.DEBUG)
    family = {"strain1": [],
              "strain2": ["member2"],
              "strain3": ["member3", "hello"],
              "strain4": ["my_member"]}
    assert not persf.uniq_members(family)
    assert "Problem, no members for strain1!" in caplog.text

def test_mixed():
    """
    Test that it returns true when there is exactly 1 member in 3 genomes/4 (and minimum
    asked is 3), and several members in the other strain
    """
    family = {"strain1": ["member"],
              "strain2": ["member2"],
              "strain3": ["member3"],
              "strain4": ["member4", "member4bis"]}
    assert persf.mixed_family(family, 3)


def test_mixed_empty(caplog):
    """
    Test that when there is exactly 1 member in 3 genomes / 4 (and min 3 asked),
    and 0 in the other genomes, it returns True.
    """
    caplog.set_level(logging.DEBUG)
    family = {"strain1": ["member"],
              "strain2": ["member2"],
              "strain2b": [],
              "strain3b": [],
              "strain4": [],
              "strain3": ["member3"],
              "strain5": []}
    assert persf.mixed_family(family, 3)
    assert "Problem, no members for strain2b" in caplog.text
    assert "Problem, no members for strain3b" in caplog.text
    assert "Problem, no members for strain4" in caplog.text


def test_not_mixed():
    """
    Test that it returns false when there are less than the required
    number of genomes with exactly 1 member
    """
    family = {"strain1": ["member"],
              "strain2": ["member2"],
              "strain3": ["member3"],
              "strain4": ["member4", "member4bis"]}
    assert not persf.mixed_family(family, 4)


def test_mixed_more():
    """
    Test that it returns true when there are more than the required
    number of genomes with exactly 1 member
    """
    family = {"strain1": ["member"],
              "strain2": ["member2"],
              "strain3": ["member3"],
              "strain4": ["member4"],
              "strain5": ["member5"]}
    assert persf.mixed_family(family, 3)


def test_write_pers():
    """
    Test that output file is written as expected
    """
    fams = {9: ["member_3", "member_12", "other_member_2"],
            3: ["member_10", "member_100", "member_1"],
            10: ["member_1", "member_2", "member_3"],
            1: ["my_protein_3", "my_protein_12", "my_protein_2"],
            5: ["ESCO.1216.00003.i001_01001", "SAEN.0215.00003.i009_00001",
                "ESCO.1017.00003.b001_00001", "ESCO.0812.00002.i002_02000",
                "ESCO.0812.00003.i002_02000"]}
    outfile = "test-persistent_families.txt"
    persf.write_persistent(fams, outfile)
    expfile = os.path.join(EXP_PATH, "exp_persgenome1.txt")
    assert tutils.compare_order_content(outfile, expfile)
    os.remove(outfile)


def test_get_core_strict(caplog):
    """
    Getting a core genome (4 genomes, all having exactly 1 member)
    """
    caplog.set_level(logging.DEBUG)
    fams = persf.get_pers(FAMS_BY_STRAIN, FAMILIES, 4)
    exp_fams = {num: mems for num, mems in FAMILIES.items() if num in ['3', '5']}
    assert exp_fams == fams
    assert ("The core genome contains 2 families, each one having "
            "exactly 4 members, from the 4 different genomes.") in caplog.text


def test_get_core_multi(caplog):
    """
    Getting a multi core genome (4 genomes, having at least 1 member)
    """
    caplog.set_level(logging.DEBUG)
    fams = persf.get_pers(FAMS_BY_STRAIN, FAMILIES, 4, multi=True)
    exp_fams = {num: mems for num, mems in FAMILIES.items() if num in ['3', '5', '1', '12']}
    assert exp_fams == fams
    assert ("The persistent genome contains 4 families with members present in "
            "at least 4 different genomes (100% of the total number of genomes)") in caplog.text


def test_get_99pers_floor_strict(caplog):
    """
    Getting a strict persistent at floor(99%) -> at least 3 genomes with 1 member, others
    absent.
    """
    caplog.set_level(logging.DEBUG)
    fams = persf.get_pers(FAMS_BY_STRAIN, FAMILIES, 4, tol=0.99, floor=True)
    exp_fams = {num: mems for num, mems in FAMILIES.items() if num in ['3', '5', '8', '10', '11']}
    assert exp_fams == fams
    assert ("The persistent genome contains 5 families, each one "
            "having exactly 1 member from at least 99.0% of the 4 different genomes "
            "(that is 3 genomes). The other genomes are absent from the family.") in caplog.text


def test_get_99pers_floor_mixed(caplog):
    """
    Getting a mixed persistent at floor(99%) -> at least 3 genomes with 1 member, others
    anything
    """
    caplog.set_level(logging.DEBUG)
    fams = persf.get_pers(FAMS_BY_STRAIN, FAMILIES, 4, tol=0.99, floor=True, mixed=True)
    exp_fams = {num: mems for num, mems in FAMILIES.items() if num in ['1', '3', '5', '8', '10',
                                                                       '11', '12']}
    assert exp_fams == fams
    assert ("The persistent genome contains 7 families, each one having exactly 1 member from at least "
            "99.0% of the genomes (3 genomes). In the remaining "
            "1.0% genomes, there can be 0, 1 or several members.") in caplog.text


def test_get_99pers_floor_multi(caplog):
    """
    Getting a multi persistent genome at floor(99%) -> at least 3 genomes (any number of members)
    """
    caplog.set_level(logging.DEBUG)
    fams = persf.get_pers(FAMS_BY_STRAIN, FAMILIES, 4, tol=0.99, floor=True, multi=True)
    exp_fams = {num: mems for num, mems in FAMILIES.items() if num in ['1', '3', '5', '6', '8',
                                                                       '10', '11', '12']}
    assert exp_fams == fams
    assert ("The persistent genome contains 8 families with members present in "
            "at least 3 different genomes (99.0% of the total number of genomes).") in caplog.text


def test_get_99pers_strict(caplog):
    """
    Getting a persistent genome at 99% (ceil) -> 4 genomes with exactly 1member
    """
    caplog.set_level(logging.DEBUG)
    fams = persf.get_pers(FAMS_BY_STRAIN, FAMILIES, 4, tol=0.99)
    exp_fams = {num: mems for num, mems in FAMILIES.items() if num in ['3', '5']}
    assert exp_fams == fams
    assert ("The persistent genome contains 2 families, each one "
            "having exactly 1 member from at least 99.0% of the 4 different genomes "
            "(that is 4 genomes). The other genomes are absent from the family.") in caplog.text


def test_get_99pers_mixed(caplog):
    """
    Getting a mixed persistent genome at 99% (ceil) -> 4 genomes with exactly 1member
    """
    caplog.set_level(logging.DEBUG)
    fams = persf.get_pers(FAMS_BY_STRAIN, FAMILIES, 4, tol=0.99, mixed=True)
    exp_fams = {num: mems for num, mems in FAMILIES.items() if num in ['3', '5']}
    assert exp_fams == fams
    assert ("The persistent genome contains 2 families, each one having exactly 1 member from at least "
            "99.0% of the genomes (4 genomes). In the remaining "
            "1.0% genomes, there can be 0, 1 or several members.") in caplog.text


def test_get_99pers_multi(caplog):
    """
    Getting a multi persistent genome at 99% (ceil) -> 3 genomes with exactly 1member,
    other with anything
    """
    caplog.set_level(logging.DEBUG)
    fams = persf.get_pers(FAMS_BY_STRAIN, FAMILIES, 4, tol=0.99, multi=True)
    exp_fams = {num: mems for num, mems in FAMILIES.items() if num in ['1', '3', '5', '12']}
    assert exp_fams == fams
    assert ("The persistent genome contains 4 families with members present in "
            "at least 4 different genomes (99.0% of the total number of genomes).") in caplog.text



#!/usr/bin/env python3
# coding: utf-8

# ###############################################################################
# This file is part of PanACOTA.                                                #
#                                                                               #
# Authors: Amandine Perrin                                                      #
# Copyright © 2018-2020 Institut Pasteur (Paris).                               #
# See the COPYRIGHT file for details.                                           #
#                                                                               #
# PanACOTA is a software providing tools for large scale bacterial comparative  #
# genomics. From a set of complete and/or draft genomes, you can:               #
#    -  Do a quality control of your strains, to eliminate poor quality         #
# genomes, which would not give any information for the comparative study       #
#    -  Uniformly annotate all genomes                                          #
#    -  Do a Pan-genome                                                         #
#    -  Do a Core or Persistent genome                                          #
#    -  Align all Core/Persistent families                                      #
#    -  Infer a phylogenetic tree from the Core/Persistent families             #
#                                                                               #
# PanACOTA is free software: you can redistribute it and/or modify it under the #
# terms of the Affero GNU General Public License as published by the Free       #
# Software Foundation, either version 3 of the License, or (at your option)     #
# any later version.                                                            #
#                                                                               #
# PanACOTA is distributed in the hope that it will be useful, but WITHOUT ANY   #
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS     #
# FOR A PARTICULAR PURPOSE. See the Affero GNU General Public License           #
# for more details.                                                             #
#                                                                               #
# You should have received a copy of the Affero GNU General Public License      #
# along with PanACOTA (COPYING file).                                           #
# If not, see <https://www.gnu.org/licenses/>.                                  #
# ###############################################################################

"""
Functions to convert prodigal result files to gembase format.

    * Proteins: multifasta with all CDS in aa
    * Replicons: (multi)fasta of genome sequences
    * Genes: multifasta of all genes in nuc
    * gff3: gff files without sequence
    * LSTINFO: information on annotation. Columns are: "start end strand type locus
    gene_name | product | EC_number | inference2" and strain is C (complement) or D (direct).
    Locus is: `<genome_name>.<contig_num><i or b>_<protein_num>`
    For prodigal: "start end strand type locus NA | NA | NA | NA", as there is no
    functional annotation.

@author gem
July 2019
"""

import os
import shutil
import glob
import logging

import PanACoTA.utils as utils
from PanACoTA.annotate_module import general_format_functions as gfunc

logger = logging.getLogger("annotate.prodigal_format")


def format_one_genome(gpath, name, prod_path, lst_dir, prot_dir, gene_dir,
                      rep_dir, gff_dir):
    """
    Format the given genome, and create its corresponding files in the following folders:

    - Proteins
    - Genes
    - Replicons
    - LSTINFO
    - gff

    Parameters
    ----------
    gpath : str
        path to the genome sequence which was given to prodigal for annotation
    name : str
        gembase name of the genome
    prod_path : str
        directory where all tmp_files for all sequences are saved (sequences cut at each set
        of 5N, prodigal results and logs)
    lst_dir : str
        path to LSTINFO folder
    prot_dir : str
        path to Proteins folder
    gene_dir : str
        path to Genes folder
    rep_dir : str
        path to Replicons folder
    gff_dir : str
        path to gff3 folder

    Returns
    -------
    bool :
        True if genome was correctly formatted, False otherwise
    """
    # Get directory where prodigal results for the current genome are saved
    prodigal_dir = os.path.join(prod_path, os.path.basename(gpath) + "-prodigalRes")

    # Get prodigal result files
    prot_file = glob.glob(os.path.join(prodigal_dir, "*.faa"))[0]
    gen_file = glob.glob(os.path.join(prodigal_dir, "*.ffn"))[0]
    gff_file = glob.glob(os.path.join(prodigal_dir, "*.gff"))[0]

    # Define names for generated gembase files
    res_prot_file = os.path.join(prot_dir, name + ".prt")
    res_gene_file = os.path.join(gene_dir, name + ".gen")
    res_lst_file = os.path.join(lst_dir, name + ".lst")
    res_rep_file = os.path.join(rep_dir, name + ".fna")
    res_gff_file = os.path.join(gff_dir, name + ".gff")

    # Generate replicon file (same as input sequence but with gembase formatted headers). From
    # this file, get contig names, to be used to generate gff file
    contigs, sizes = utils.get_genome_contigs_and_rename(name, gpath, res_rep_file, logger)
    if not contigs:
        try:
            os.remove(res_rep_file)
            os.remove(res_lst_file)
            os.remove(res_gff_file)
            os.remove(res_gene_file)
            os.remove(res_prot_file)
        except OSError:
            pass
        logger.error("Problems while generating Replicon file for {}".format(name))
        return False

    # First, create .gen and .lst files. If they could not be formatted,
    # remove those files, and return False with error message
    ok = create_gene_lst(contigs, gen_file, res_gene_file, res_lst_file, gpath, name)
    if not ok:
        try:
            os.remove(res_rep_file)
            os.remove(res_gene_file)
            os.remove(res_lst_file)
            os.remove(res_gff_file)
            os.remove(res_prot_file)
        except OSError:
            pass
        logger.error(f"Problems while generating .gen and .lst files for {name}")
        return False

    # Create gff files.
    ok = create_gff(gpath, gff_file, res_gff_file, res_lst_file, contigs, sizes)
    # If problem while formatting the genome (rep or gff file), remove all
    # already created files, and return False (genome not formatted) with error message.
    if not ok:
        try:
            os.remove(res_gene_file)
            os.remove(res_lst_file)
            os.remove(res_rep_file)
            os.remove(res_gff_file)
            os.remove(res_prot_file)
        except OSError:
            pass
        logger.error("Problems while generating .gff (gff3 folder) "
                     "file for {}".format(name))
        return False

    # Generate .prt files (in Proteins directory)
    ok = create_prt(prot_file, res_prot_file, res_lst_file)
    # If problem while formatting prt file, return False, delete all generated
    # formatted files, and write an error message to user.
    if not ok:
        try:
            os.remove(res_gene_file)
            os.remove(res_lst_file)
            os.remove(res_rep_file)
            os.remove(res_gff_file)
            os.remove(res_prot_file)
        except OSError: # pragma: no cover
            pass
        logger.error("Problems while generating .prt file (Proteins folder) "
                     "for {}".format(name))
        return False
    return ok


def create_gene_lst(contigs, gen_file, res_gen_file, res_lst_file, gpath, name):
    """
    Generate .gen file, from sequences contained in .ffn, but changing the
    headers to match with gembase format.
    At the same time, generate .lst file, from the information given in prodigal ffn headers

    Parameters
    ----------
    contigs : dict
        {original_contig_name: gembase_contig_name}
    gen_file : str
        .ffn file generated by prodigal
    res_gen_file : str
        generated .gen file, to write in Genes directory
    res_lst_file : str
        generated .lst file to write in LSTINFO directory
    gpath : str
        path to the genome given to prodigal. Only used for error message
    name : str
        gembase name of the genome to format
    logger : logging.Logger
        logger object to put information

    Returns
    -------
    bool :
        True if conversion went well, False otherwise
    """
    # Variable which will contain the current gene sequence
    seq = ""
    # number of the current gene (first gene is 1, 2nd is 2 etc. each number is unique: do not
    # re-start at 1 for each new contig)
    locus_num = 0
    # contig name of the last gene. To check if we are now in a new contig (-> loc = b) or not
    prev_cont_name = ""
    # Previous ontig number: contig number to use in gembase format
    prev_cont_num = 0
    contig_num = 0
    # Keep start, end, strand and informations (prodigal gives information on start_type,
    # gc_cont etc.) from the previous gene, before overwriting it with information
    # on the new one
    prev_start = ""
    prev_end = ""
    prev_strand = ""
    prev_info = ""
    # Update loc when contig changes ('b' if gene at the border of a contig, 'i' if it is inside)
    prev_loc = "b"
    # To start, the first gene is, by definition, at the border of the contig
    loc = "b"
    # Open files: .ffn prodigal to read, .gen and .lst gembase to create
    with open(gen_file, "r") as ffn, open(res_gen_file, "w") as r_gen,\
         open(res_lst_file, "w") as r_lst:
        # Read all lines in ffn file (sequences in nuc. for each gene)
        for lineffn in ffn:
            # If it is a sequence, save it and go to next line
            if not lineffn.startswith(">"):
                seq += lineffn
                continue
            # Otherwise:
            # - write header of previous sequence to .gen
            # - write previous sequence (in 'seq') to .gen
            # - write LSTINFO information to .lst
            # - update information (new start, end, contig number etc.) for next gene
            else:
                # Get information given for the new gene (by .ffn file from prodigal)
                (gname, start, end, strand, info) = lineffn.strip().split(">")[-1].split("#")
                # Get contig number from prodigal gene header: prodigal first part of header is:
                #  <original genome name contig name>_<protein number>
                contig_name = gname.strip().split("_")
                if len(contig_name) > 1:
                    contig_name = "_".join(contig_name[:-1])
                else:
                    contig_name = contig_name[0]
                # If new contig:
                # - previous gene was the last of its contig -> prev_loc = "b" ;
                # - the current gene is the first of its contig (loc = "b")
                # - we must increment the contig number
                if contig_name != prev_cont_name:
                    # Check that this contig name is in the list, and get its gembase contig number
                    if contig_name in contigs:
                        contig_num = contigs[contig_name].split(".")[-1]
                    # if not in the list, problem, return false
                    else:
                        logger.error(f"'{contig_name}' found in {gen_file} does not exist in "
                                     f"{gpath}.")
                        return False
                    prev_loc = 'b'
                    loc = 'b'
                # If not new contig. If prev_loc == 'b', previous gene is the first protein
                # of this contig.
                # Current gene will be inside the contig (except if new contig for the next gene,
                # meaning only 1 gene in the contig)
                else:
                    loc = 'i'

                # If it is not the first gene of the genome, write previous gene information
                if prev_start != "":
                    # Write line in LSTINFO file, + header and sequence to the gene file
                    lstline = gfunc.write_gene("CDS", locus_num, "NA", "NA",
                                               prev_loc, name, prev_cont_num, "NA", prev_info,
                                               "NA", prev_strand, prev_start, prev_end, r_lst)
                    gfunc.write_header(lstline, r_gen)
                    r_gen.write(seq)
                # -> get new information, save it for the next gene, and go to next line
                # Strands are 1/-1 in prodigal, while we use D,C -> convert, so that next time
                # we find a new gene, it writes this before updating for this new gene
                if int(strand) == 1:
                    strand = "D"
                else:
                    strand = "C"
                # Prepare variables for next gene
                locus_num += 1
                seq = ""
                prev_cont_num = contig_num
                prev_cont_name = contig_name
                prev_start = start
                prev_end = end
                prev_strand = strand
                prev_loc = loc
                prev_info = info
        # Write last gene of the genome (-> loc = 'b'),
        # Just check that there was at least 1 gene found (prev_start != "").
        # Otherwise, nothing to write
        if prev_start != "":
            prev_loc = "b"
            lstline = gfunc.write_gene("CDS", locus_num, "NA", "NA",
                                       prev_loc, name, prev_cont_num, "NA", prev_info, "NA",
                                       prev_strand, prev_start, prev_end, r_lst)
            gfunc.write_header(lstline, r_gen)
            r_gen.write(seq)
    return True


def create_gff(gpath, gff_file, res_gff_file, res_lst_file, contigs, sizes):
    """
    Create .gff3 file.

    Format:

    ##gff-version 3
    ##sequence-region contig1 start end
    ##sequence-region contig2 start end
    ...
    seqid(=contig) source   type start   end score  strand phase attributes

    All fields tab separated.
    Empty fields contain '.'

    For example:
    ESCO.1017.00200.00001    Prodigal:2.6    CDS start  end .   +   .   ID=ESCO.1017.00200.b0001_00001;locus_tag=ESCO.1017.00200.b0001_00001;product=hypothetical protein


    genome1_1   Prodigal_v2.6.3 CDS 213 1880    260.0   +   0   ID=1_1;partial=00;start_type=ATG;rbs_motif=None;rbs_spacer=None;gc_cont=0.534;conf=99.99;score=259.99;cscore=268.89;sscore=-8.89;rscore=-8.50;uscore=-4.34;tscore=3.95;

    Parameters
    ----------
        gpath : str
            path to the genome sequence given to prodigal. Only used for the error message
        gff_file : str
            path to gff file generated by prodigal
        res-gff_file : str
            path to the gff file that must be created in result database
        res-lst_file : str
            path to the lst file that was created in result database in the previous step
        contigs : dict
            dict of contig names with their size. ["original_name": "gembase_name"]
        sizes : dict
            dict of contig gembase names with their sizes {"gembase_name": size}

    Returns
    -------
    bool
        True if everything went well, False if any problem

    """
    # open gff generated by prodigal to read it
    # open file to write new gff file
    # open lst file to read all information saved from prodigal results
    with open(gff_file, 'r') as gf, open(res_gff_file, "w") as rgf, open(res_lst_file, "r") as rlf:
        # Write headers of gff3 file
        rgf.write("##gff-version  3\n")
        for ori_name, new_name in contigs.items():
            # Write the list of contigs, with their size
            end = sizes[new_name]
            rgf.write(f"##sequence-region\t{new_name}\t1\t{end}\n")

        # Now, convert each line of prodigal gff to gembase formatted gff line
        for linegff in gf:
            # Ignore gff lines starting by #. For the new gff, these are already all written at the
            # beginning of the file.
            if linegff.startswith("#"):
                continue

            # We do not write the sequences
            if linegff.startswith(">"):
                break

            # Get all information from prodigal gff line. Strip each of them as trailing whitespace
            # can be hidden (leading to information from gff considered as different from
            # information from lst)
            fields_g = linegff.split("\t")
            fields_g = [info.strip() for info in fields_g]
            (contig_name, source, type_g, start_g, end_g,
             score, strand_g, phase, attributes) = fields_g
            # Get information given to this same sequence from the lst file
            # (next lst line corresponds to next gff line without #), as, for each format,
            # there is 1 line per gene)
            linelst = rlf.readline()
            fields_l = linelst.split("\t")
            fields_l = [info.strip() for info in fields_l]
            start_l, end_l, strand_l, type_l, locus_l, _, _ = fields_l


            # Get gff and ffn filenames to give information to user if error message
            gff = os.path.basename(gff_file)
            ffn = ".".join(gff.split(".")[:-1]) + ".ffn"
            # Path where gff and ffn generated by prodigal are
            tmp = gpath + "-prodigalRes"
            # Get gene name given by prodigal to current gene
            gname = attributes.split("ID=")[1].split(";")[0]

            # Compare information from lst and information from prodigal gff (start,
            # end and type of feature). They should correspond
            for (elemg, eleml, label) in zip([start_g, end_g, type_g],
                                             [start_l, end_l, type_l],
                                             ["start", "end", "type"]):
                # If 1 element is different (start or end position, or type), print error
                # message and return False: this genome could not be converted to gff
                if elemg != eleml:
                    logger.error(f"Files {ffn} and {gff} (in prodigal tmp_files: {tmp}) do not have "
                                 f"the same {label} value for gene {gname} ({elemg} in gff, {eleml} in "
                                 "ffn))")
                    return False

            # Replace prodigal ID with the new gene name (in gembase format), found in the
            # corresponding lst line
            at_split = attributes.split(";")
            new = [atr if "ID" not in atr else f'ID={locus_l}' for atr in at_split]

            # Write new line of gff file
            # Write new contig name
            cname = contigs[contig_name]
            info = "\t".join([cname, source, type_g, start_g, end_g, score, strand_g,
                              phase, ";".join(new)])
            rgf.write(info + "\n")
    return True


def create_prt(prot_file, res_prot_file, res_lst_file):
    """
    Generate .prt file (gembase formatted gene names), from features contained in .lst file generated just before.

    Parameters
    ----------
    prot_file : str
        .faa file generated by prodigal
    res_prot_file : str
        output file, to write in Proteins directory
    res_lst_file : str
        .lst file to get all gene names in gembase format instead of re-generating them
    Returns
    -------
    bool :
        True if conversion went well, False otherwise
    """

    # Open:
    # - prot file to read gene sequences from prodigal results
    # - res_prot file to write sequences with gembase headers
    # - res_lst_file to get gene gembase names and other infos (strand, size...)

    with open(prot_file, "r") as faa, open(res_prot_file, "w") as r_prt,\
         open(res_lst_file, "r") as r_lst:
         # Read prt file generated by prodigal
        for lineprot in faa:
            # If protein sequence, write it
            if not lineprot.startswith(">"):
                r_prt.write(lineprot)
                continue
            # If header, replace by gembase header
            # For that, get next lst line (corresponding to next protein,
            # as there is 1 protein per line in .lst -> 1 protein per header in .prt)
            linelst = r_lst.readline().strip()
            # Try to get info from lstline.
            # If lstline empty, it means that the current protein
            # is missing from lst file. We already read the last protein of lst file.
            if linelst != '':
                # If not empty, check lst format, return False if not right format
                try:
                    # If ok, gembase name is in the fifth column of lst file
                    start, end, _, _, gem_name, _, _ = linelst.split("\t")
                except ValueError:
                    logger.error("Problem in format of lstline ({})".format(linelst))
                    return False
            else:
                logger.error("No more protein in lst file. We cannot get information on this "
                             "protein ({})! Check that you do not have more proteins than genes "
                             "in prodigal results".format(lineprot.strip()))
                return False
            # Write this gembase name as a new header
            # Size of protein sequence is the third of gene sequence. Check that it is an int.
            try :
                size_gen = (int(end) - int(start) + 1)
            except ValueError:
                logger.error("Start and/or end of protein {} position is not a number (start "
                             "= {}; end = {})".format(gem_name, start, end))
                return False
            # Find size of protein in number of aa
            # If number of nucleotides in gene cannot be divided by 3 to get number of amino-acids, there is a problem with this protein: return False to ignore this genome
            size_prot = size_gen/3
            if int(size_prot) != size_prot:
                logger.error("Gene {} has a number of nucleotides ({}) that is not divisible "
                             "by 3.".format(gem_name, size_gen))
                return False
            gfunc.write_header(linelst, r_prt)
            # new_header = "\t".join([gem_name, str(int(size_prot)), product, info])
            # r_prt.write(">" + new_header + "\n")
        # Check that there are no more proteins in lst than in this prt file
        linelst = r_lst.readline()
        if linelst.strip() != '':
            gem_name = linelst.strip().split("\t")[4]
            logger.error("Protein {} is in .lst file but its sequence is not in the protein "
                         "file generated by prodigal.".format(gem_name))
            return False
    return True



#!/usr/bin/env python3
# coding: utf-8

"""
Unit tests for the alignment submodule in align module
"""
import logging
import os
import subprocess
import pytest
import shutil

import multiprocessing

import PanACoTA.align_module.alignment as al
import test.test_unit.utilities_for_tests as tutil


# Define common variables
ALDIR = os.path.join("test", "data", "align")
EXPPATH = os.path.join(ALDIR, "exp_files")
TESTPATH = os.path.join(ALDIR, "test_files")
GENEPATH = os.path.join(ALDIR, "generated_by_unit-tests")


@pytest.fixture(autouse=True)
def setup_teardown_module():
    """
    Remove log files at the end of this test module

    Before each test:
    - init logger
    - create directory to put generated files

    After:
    - remove all log files
    - remove directory with generated results
    """
    os.mkdir(GENEPATH)
    print("setup")

    yield
    shutil.rmtree(GENEPATH)
    print("teardown")



def test_check_len():
    """
    Test that when giving an alignment file with all sequences having the same length,
    it returns this length, and the number of sequences aligned
    """
    aln_file = os.path.join(TESTPATH, "test_alignment-mafft-ok.aln")
    logger = logging.getLogger("test_alignment_check_len")
    fam = 1
    res = al.check_lens(aln_file, fam, logger)
    assert res[0] == 157
    assert res[1] == 4


def test_check_len_diff(caplog):
    """
    Test that when giving an alignment file with sequences not having all the same length,
    it returns False, and an error message
    """
    caplog.set_level(logging.DEBUG)
    aln_file = os.path.join(TESTPATH, "test_alignment-mafft-diff-len.aln")
    logger = logging.getLogger("test_alignment_check_len")
    fam = 1
    assert not al.check_lens(aln_file, fam, logger)
    assert ("alignments for family 1 do not all have the same length. "
            "Lengths found are: {") in caplog.text
    assert "126" in caplog.text
    assert "154" in caplog.text
    assert "157" in caplog.text
    assert "}\n" in caplog.text


def test_check_btr():
    """
    Test that when giving a nucleotide alignment, and a number of sequences corresponding to the
    number of sequences in the alignment file, it returns the number of sequences.
    """
    btr_file = os.path.join(TESTPATH, "test_alignment-btr-ok.aln")
    nbfal = 4
    logger = logging.getLogger("test_check_btr")
    message = "problem!"
    assert al.check_nb_seqs(btr_file, nbfal, logger, message) == 4


def test_check_btr_listsizes():
    """
    Test that when giving a nucleotide alignment, and a list of numbers, including 1
    corresponding to the number of sequences corresponding to the
    number of sequences in the alignment file, it returns the number of sequences in the
    alignment file.
    """
    btr_file = os.path.join(TESTPATH, "test_alignment-btr-ok.aln")
    nbfal = [3, 4, 5]
    logger = logging.getLogger("test_check_btr")
    message = "problem!"
    assert al.check_nb_seqs(btr_file, nbfal, logger, message) == 4


def test_check_btr_wrongnbfam(caplog):
    """
    Test that when giving a nucleotide alignment, and a number of sequences different from the
    number of sequences in the alignment file, it returns False, and an error message.
    """
    caplog.set_level(logging.DEBUG)
    btr_file = os.path.join(TESTPATH, "test_alignment-btr-ok.aln")
    nbfal = 5
    logger = logging.getLogger("test_check_btr")
    message = "Problem with fam 1"
    assert not al.check_nb_seqs(btr_file, nbfal, logger, message)
    assert "Problem with fam 1 (4)" in caplog.text


def test_check_btr_listwrongnbfam(caplog):
    """
    Test that when giving a nucleotide alignment, and a number of sequences different from the
    number of sequences in the alignment file, it returns False, and an error message.
    """
    caplog.set_level(logging.DEBUG)
    btr_file = os.path.join(TESTPATH, "test_alignment-btr-ok.aln")
    nbfal = [3, 5, 7]
    logger = logging.getLogger("test_check_btr")
    message = "Problem with fam 1"
    assert not al.check_nb_seqs(btr_file, nbfal, logger, message)
    assert "Problem with fam 1 (4)" in caplog.text


def test_backtranslate(caplog):
    """
    Test that when giving an alignment file and a file with nucleic sequences, it generates a
    btr file, with the nucleotide alignments as expected, and returns True
    """
    caplog.set_level(logging.DEBUG)
    num_fam = 1
    mafft_file = os.path.join(TESTPATH, "test_alignment-mafft-ok.aln")
    gen_file = os.path.join(EXPPATH, "exp_aldir", "current.1.gen")
    btr_file = os.path.join(GENEPATH, "test_btr_output.aln")
    nbfal = 4
    logger = logging.getLogger("test_backtranslate")
    # check that it returns true
    assert al.back_translate(num_fam, mafft_file, gen_file, btr_file, nbfal, logger) == 4
    # Check logs
    assert "Back-translating family 1" in caplog.text
    # Chek btr file content
    assert os.path.isfile(btr_file)
    exp_btr = os.path.join(EXPPATH, "exp_btr.1.aln")
    assert tutil.compare_file_content(btr_file, exp_btr)


def test_backtranslate_wrongnbfam(caplog):
    """
    Test that when giving an alignment file and a file with nucleic sequences, it generates a
    btr file, with the nucleotide alignments as expected, but if we give a wrong number of
    expected sequences in btr file, it returns False with an error message
    """
    caplog.set_level(logging.DEBUG)
    num_fam = 1
    mafft_file = os.path.join(TESTPATH, "test_alignment-mafft-ok.aln")
    gen_file = os.path.join(EXPPATH, "exp_aldir", "current.1.gen")
    btr_file = os.path.join(GENEPATH, "test_btr_output.aln")
    nbfal = 3
    logger = logging.getLogger("test_backtranslate")
    # check that it returns true
    assert not al.back_translate(num_fam, mafft_file, gen_file, btr_file, nbfal, logger)
    # Check logs
    assert "Back-translating family 1" in caplog.text
    assert ("fam 1: different number of proteins aligned in "
            "test/data/align/test_files/test_alignment-mafft-ok.aln (3) and genes back-translated in "
            "test/data/align/generated_by_unit-tests/test_btr_output.aln ") in caplog.text
    # Chek btr file content
    assert os.path.isfile(btr_file)
    exp_btr = os.path.join(EXPPATH, "exp_btr.1.aln")
    assert tutil.compare_file_content(btr_file, exp_btr)


def test_backtranslate_problem(caplog):
    """
    Test that when giving an alignment file and a wrong filename for the  nucleic sequences,
    it generates an empty btr file, and returns False, and an error message.
    """
    caplog.set_level(logging.DEBUG)
    num_fam = 1
    mafft_file = os.path.join(TESTPATH, "test_alignment-mafft-ok.aln")
    gen_file = os.path.join(EXPPATH, "exp_aldir", "current.1")
    btr_file = os.path.join(GENEPATH, "test_btr_output.aln")
    nbfal = 4
    logger = logging.getLogger("test_backtranslate")
    # check that it returns true
    assert not al.back_translate(num_fam, mafft_file, gen_file, btr_file, nbfal, logger)
    # Check logs
    assert "Back-translating family 1" in caplog.text
    assert ("Problem while trying to backtranslate test/data/align/test_files/test_alignment-"
            "mafft-ok.aln to a nucleotide alignment") in caplog.text
    # Chek btr file content
    assert not os.path.isfile(btr_file)


def test_check_mafft_align():
    """
    Test that when giving an alignment file, and a number of sequences equal to the number of
    sequences in the alignment file, it returns this number.
    """
    nbfal = 4
    mafft_file = os.path.join(TESTPATH, "test_alignment-mafft-ok.aln")
    message = "problem!"
    logger = logging.getLogger("test_check_mafft_align")
    assert al.check_nb_seqs(mafft_file, nbfal, logger, message) == 4


def test_mafft_align(caplog):
    """
    Test that when giving a file containing extracted proteins, it aligns them as expected
    and returns True
    """
    caplog.set_level(logging.DEBUG)
    num_fam = 1
    prt_file = os.path.join(EXPPATH, "exp_aldir", "current.1.prt")
    mafft_file = os.path.join(GENEPATH, "test_mafft_align.aln")
    nbfprt = 4
    logger = logging.getLogger("test_check_mafft_align")
    assert al.mafft_align(num_fam, prt_file, mafft_file, nbfprt, logger) == 4
    assert "Aligning family 1" in caplog.text
    assert os.path.isfile(mafft_file)
    exp_mafft = os.path.join(EXPPATH, "exp_aldir", "mafft-align.1.aln")
    assert tutil.compare_file_content(mafft_file, exp_mafft)


def test_mafft_align_error(caplog):
    """
    Test that when giving a wrong file with the sequence to extract (non existing file),
    it returns false with the expected error message
    """
    caplog.set_level(logging.DEBUG)
    num_fam = 1
    prt_file = "prtfile"
    mafft_file = os.path.join(GENEPATH, "test_mafft_align.aln")
    nbfprt = 4
    logger = logging.getLogger("test_check_mafft_align")
    assert not al.mafft_align(num_fam, prt_file, mafft_file, nbfprt, logger)
    assert "Aligning family 1" in caplog.text
    assert "Problem while trying to align fam 1" in caplog.text
    assert not os.path.isfile(mafft_file)


def test_mafft_align_wrongnbfam(caplog):
    """
    Test that when giving a file containing extracted protein to align, it aligns them,
    but as we don't give the same number of expected aligned proteins, it returns False,
    with an error message
    """
    caplog.set_level(logging.DEBUG)
    num_fam = 1
    prt_file = os.path.join(EXPPATH, "exp_aldir", "current.1.prt")
    mafft_file = os.path.join(GENEPATH, "test_mafft_align.aln")
    nbfprt = 3
    logger = logging.getLogger("test_check_mafft_align")
    assert not al.mafft_align(num_fam, prt_file, mafft_file, nbfprt, logger)
    assert "Aligning family 1" in caplog.text
    assert ("fam 1: different number of proteins extracted in "
            "test/data/align/exp_files/exp_aldir/current.1.prt (3) and proteins aligned "
            "in test/data/align/generated_by_unit-tests/test_mafft_align.aln (4)") in caplog.text
    assert os.path.isfile(mafft_file)
    exp_mafft = os.path.join(EXPPATH, "exp_aldir", "mafft-align.1.aln")
    assert tutil.compare_file_content(mafft_file, exp_mafft)


def test_check_extract(caplog):
    """
    Test that given the 3 files (prt, gen and miss) for a given family, having the expected number
    of sequences, it returns True
    """
    caplog.set_level(logging.DEBUG)
    num_fam = 1
    gen_file = os.path.join(EXPPATH, "exp_aldir", "current.1.gen")
    prt_file = os.path.join(EXPPATH, "exp_aldir", "current.1.prt")
    miss_file = os.path.join(GENEPATH, "test_check_extract_miss-file.txt")
    ngenomes = 5
    logger = logging.getLogger("test_check_extract")
    with open(miss_file, "w") as missf:
        missf.write("Genome5")
    assert al.check_extractions(num_fam, miss_file, prt_file, gen_file, ngenomes, logger) == 4
    assert "Checking extractions for family 1" in caplog.text


def test_check_extract_wrongnbmiss(caplog):
    """
    Test that given the 3 files: 4 proteins extracted in gen and prt, empty miss file,
    and saying that there are 5 genomes in the dataset, it returns False, with an error message
    indicating the wrong sum of prt_seqs + miss (4 instead of 5 expected)
    """
    caplog.set_level(logging.DEBUG)
    num_fam = 1
    gen_file = os.path.join(EXPPATH, "exp_aldir", "current.1.gen")
    prt_file = os.path.join(EXPPATH, "exp_aldir", "current.1.prt")
    miss_file = os.path.join(GENEPATH, "test_check_extract_miss-file.txt")
    ngenomes = 5
    logger = logging.getLogger("test_check_extract")
    open(miss_file, "w").close()
    assert not al.check_extractions(num_fam, miss_file, prt_file, gen_file, ngenomes, logger)
    assert "Checking extractions for family 1" in caplog.text
    assert ("fam 1: wrong sum of missing genomes (0) and prt extracted (4) for 5 genomes in "
            "the dataset.") in caplog.text


def test_check_extract_wrongnbgen(caplog):
    """
    Test that given the 3 files: 4 proteins extracted in gen, 5 in prt, empty miss file,
    and saying that there are 5 genomes in the dataset, it returns False, with an error message
    indicating the wrong sum of gen_seqs + miss (4 instead of 5 expected)
    """
    caplog.set_level(logging.DEBUG)
    num_fam = 1
    gen_file = os.path.join(EXPPATH, "exp_aldir", "current.1.gen")
    prt_file_ref = os.path.join(EXPPATH, "exp_aldir", "current.1.prt")
    prt_file = os.path.join(GENEPATH, "test_check_extract_prt-file.prt")
    shutil.copyfile(prt_file_ref, prt_file)
    miss_file = os.path.join(GENEPATH, "test_check_extract_miss-file.txt")
    ngenomes = 5
    logger = logging.getLogger("test_check_extract")
    # Create empty miss file
    open(miss_file, "w").close()
    # Add a 5th sequence in prt file
    with open(prt_file, "a") as prtf:
        prtf.write(">fifth_sequence\nATTCGC\n")
    assert not al.check_extractions(num_fam, miss_file, prt_file, gen_file, ngenomes, logger)
    assert "Checking extractions for family 1" in caplog.text
    assert ("fam 1: wrong sum of missing genomes (0) and gen extracted (4) for 5 genomes in "
            "the dataset.") in caplog.text


def test_family_align(caplog):
    """
    Test that when giving prt file (3 extracted), gen file (3 extracted), miss file (1)
    and total nb genomes = 4, it aligns all families, and backtranslates the alignment to
    nucleotides as expected, and returns the number of sequences aligned.
    """
    caplog.set_level(logging.DEBUG)
    prt_file = os.path.join(EXPPATH, "exp_aldir-pers", "current.8.prt")
    gen_file = os.path.join(EXPPATH, "exp_aldir-pers", "current.8.gen")
    miss_file = os.path.join(EXPPATH, "exp_aldir-pers", "current.8.miss.lst")
    mafft_file = os.path.join(GENEPATH, "test_fam_align.8.aln")
    btr_file = os.path.join(GENEPATH, "test_fam_align_btr.8.aln")
    num_fam = 8
    ngenomes = 4
    logger = logging.getLogger("test_fam_align")
    assert al.family_alignment(prt_file, gen_file, miss_file, mafft_file, btr_file, num_fam,
                               ngenomes, logger) == 3
    assert "Checking extractions for family 8" in caplog.text
    assert "Aligning family 8" in caplog.text
    assert "Back-translating family 8" in caplog.text
    # Check content of mafft and btr files
    exp_mafft = os.path.join(EXPPATH, "exp_aldir-pers", "mafft-align.8.aln")
    assert tutil.compare_file_content(mafft_file, exp_mafft)
    exp_btr = os.path.join(EXPPATH, "exp_aldir-pers", "mafft-btr.8.aln")
    assert tutil.compare_file_content(btr_file, exp_btr)


def test_family_align_wrongextract(caplog):
    """
    Test that when giving prt file (3 extracted), gen file (3 extracted), miss file (1)
    and total nb genomes = 5, it returns False, error message for wrong sum of prt + miss,
    and removes existing mafft and btr files.
    """
    caplog.set_level(logging.DEBUG)
    prt_file = os.path.join(EXPPATH, "exp_aldir-pers", "current.8.prt")
    gen_file = os.path.join(EXPPATH, "exp_aldir-pers", "current.8.gen")
    miss_file = os.path.join(EXPPATH, "exp_aldir-pers", "current.8.miss.lst")
    mafft_file = os.path.join(GENEPATH, "test_fam_align.8.aln")
    btr_file = os.path.join(GENEPATH, "test_fam_align_btr.8.aln")
    open(mafft_file, "w").close()
    open(btr_file, "w").close()
    num_fam = 8
    ngenomes = 5
    logger = logging.getLogger("test_fam_align")
    assert al.family_alignment(prt_file, gen_file, miss_file, mafft_file, btr_file, num_fam,
                               ngenomes, logger) is False
    assert "Checking extractions for family 8" in caplog.text
    assert ("fam 8: wrong sum of missing genomes (1) and prt extracted (3) for 5 genomes in "
            "the dataset") in caplog.text
    # Check that mafft and btr files were removed
    assert not os.path.isfile(mafft_file)
    assert not os.path.isfile(btr_file)


def test_family_align_mafftok_nobtr(caplog):
    """
    Test that when giving prt file (3 extracted), gen file (3 extracted), miss file (1)
    and total nb genomes = 4, and the alignment file with the 3 sequences aligned,
    it back-translates it and returns 3, the number of sequences back-translated
    """
    caplog.set_level(logging.DEBUG)
    prt_file = os.path.join(EXPPATH, "exp_aldir-pers", "current.8.prt")
    gen_file = os.path.join(EXPPATH, "exp_aldir-pers", "current.8.gen")
    miss_file = os.path.join(EXPPATH, "exp_aldir-pers", "current.8.miss.lst")
    # Create mafft file with all alignments
    btr_file = os.path.join(GENEPATH, "test_fam_align_btr.8.aln")
    mafft_file = os.path.join(GENEPATH, "test_fam_align.8.aln")
    ref_mafft_file = os.path.join(EXPPATH, "exp_aldir-pers", "mafft-align.8.aln")
    shutil.copyfile(ref_mafft_file, mafft_file)
    num_fam = 8
    ngenomes = 4
    logger = logging.getLogger("test_fam_align")
    assert al.family_alignment(prt_file, gen_file, miss_file, mafft_file, btr_file, num_fam,
                               ngenomes, logger) == 3
    assert "Checking extractions for family 8" in caplog.text
    assert "Back-translating family 8" in caplog.text
    # Check content of mafft and btr files
    exp_mafft = os.path.join(EXPPATH, "exp_aldir-pers", "mafft-align.8.aln")
    assert tutil.compare_file_content(mafft_file, exp_mafft)
    exp_btr = os.path.join(EXPPATH, "exp_aldir-pers", "mafft-btr.8.aln")
    assert tutil.compare_file_content(btr_file, exp_btr)


def test_family_align_mafftempty_btrempty(caplog):
    """
    Test that when giving prt file (3 extracted), gen file (3 extracted), miss file (1)
    and total nb genomes = 4, and the alignment file with the 3 sequences aligned,
    it back-translates it and returns 3, the number of sequences back-translated
    """
    caplog.set_level(logging.DEBUG)
    prt_file = os.path.join(EXPPATH, "exp_aldir-pers", "current.8.prt")
    gen_file = os.path.join(EXPPATH, "exp_aldir-pers", "current.8.gen")
    miss_file = os.path.join(EXPPATH, "exp_aldir-pers", "current.8.miss.lst")
    # Create empty mafft file
    mafft_file = os.path.join(GENEPATH, "test_fam_align.8.aln")
    open(mafft_file, "w").close()
    # Create empty btr
    btr_file = os.path.join(GENEPATH, "test_fam_align_btr.8.aln")
    open(btr_file, "w").close()
    num_fam = 8
    ngenomes = 4
    logger = logging.getLogger("test_fam_align")
    assert al.family_alignment(prt_file, gen_file, miss_file, mafft_file, btr_file, num_fam,
                               ngenomes, logger) == 3
    assert "Checking extractions for family 8" in caplog.text
    assert ("fam 8: Will redo alignment, because found a different number of proteins extracted "
            "in test/data/align/exp_files/exp_aldir-pers/current.8.prt (3) and proteins aligned "
            "in existing test/data/align/generated_by_unit-tests/test_fam_align.8.aln (0)") in caplog.text
    assert "Aligning family 8" in caplog.text
    assert "Back-translating family 8" in caplog.text
    # Check content of mafft and btr files
    exp_mafft = os.path.join(EXPPATH, "exp_aldir-pers", "mafft-align.8.aln")
    assert tutil.compare_file_content(mafft_file, exp_mafft)
    exp_btr = os.path.join(EXPPATH, "exp_aldir-pers", "mafft-btr.8.aln")
    assert tutil.compare_file_content(btr_file, exp_btr)


def test_family_align_mafftok_emptybtr(caplog):
    """
    Test that when giving prt file (3 extracted), gen file (3 extracted), miss file (1)
    and total nb genomes = 4, and the alignment file with the 3 sequences aligned,
    and an empty btr file, it re-does the backtranslation, and returns True
    """
    caplog.set_level(logging.DEBUG)
    prt_file = os.path.join(EXPPATH, "exp_aldir-pers", "current.8.prt")
    gen_file = os.path.join(EXPPATH, "exp_aldir-pers", "current.8.gen")
    miss_file = os.path.join(EXPPATH, "exp_aldir-pers", "current.8.miss.lst")
    # Create mafft file with all alignments
    mafft_file = os.path.join(GENEPATH, "test_fam_align.8.aln")
    ref_mafft_file = os.path.join(EXPPATH, "exp_aldir-pers", "mafft-align.8.aln")
    shutil.copyfile(ref_mafft_file, mafft_file)
    # Create empty btr file
    btr_file = os.path.join(GENEPATH, "test_fam_align_btr.8.aln")
    open(btr_file, "w").close()
    num_fam = 8
    ngenomes = 4
    logger = logging.getLogger("test_fam_align")
    assert al.family_alignment(prt_file, gen_file, miss_file, mafft_file, btr_file, num_fam,
                               ngenomes, logger) == 3
    assert "Checking extractions for family 8" in caplog.text
    # Check content of mafft and btr files
    exp_mafft = os.path.join(EXPPATH, "exp_aldir-pers", "mafft-align.8.aln")
    exp_btr = os.path.join(EXPPATH, "exp_aldir-pers", "mafft-btr.8.aln")
    assert tutil.compare_file_content(mafft_file, exp_mafft)
    assert tutil.compare_file_content(btr_file, exp_btr)


def test_family_align_mafftok_btrok(caplog):
    """
    Test that when giving prt file (3 extracted), gen file (3 extracted), miss file (1)
    and total nb genomes = 4, and the alignment file with the 3 sequences aligned,
    and the already generated back-translated file, it returns "OK" as it did not generate any
    new file.
    """
    caplog.set_level(logging.DEBUG)
    prt_file = os.path.join(EXPPATH, "exp_aldir-pers", "current.8.prt")
    gen_file = os.path.join(EXPPATH, "exp_aldir-pers", "current.8.gen")
    miss_file = os.path.join(EXPPATH, "exp_aldir-pers", "current.8.miss.lst")
    # Create mafft file with all alignments
    mafft_file = os.path.join(GENEPATH, "test_fam_align.8.aln")
    ref_mafft_file = os.path.join(EXPPATH, "exp_aldir-pers", "mafft-align.8.aln")
    shutil.copyfile(ref_mafft_file, mafft_file)
    # Create btr file with alignments
    btr_file = os.path.join(GENEPATH, "test_fam_align_btr.8.aln")
    ref_btr_file = os.path.join(EXPPATH, "exp_aldir-pers", "mafft-btr.8.aln")
    shutil.copyfile(ref_btr_file, btr_file)
    num_fam = 8
    ngenomes = 4
    logger = logging.getLogger("test_fam_align")
    assert al.family_alignment(prt_file, gen_file, miss_file, mafft_file, btr_file, num_fam,
                               ngenomes, logger) == "OK"
    assert "Checking extractions for family 8" in caplog.text
    # Check content of mafft and btr files
    exp_mafft = os.path.join(EXPPATH, "exp_aldir-pers", "mafft-align.8.aln")
    assert tutil.compare_file_content(mafft_file, exp_mafft)
    exp_btr = os.path.join(EXPPATH, "exp_aldir-pers", "mafft-btr.8.aln")
    assert tutil.compare_file_content(btr_file, exp_btr)


def test_family_align_nomafft_btrempty_errormafft(caplog):
    """
    Test that when giving prt file (3 extracted), gen file (3 extracted), miss file (1)
    and total nb genomes = 4, no alignment file, and an empty btr file, and make changes in
    path so that mafft crashes -> should return False, and remove btr and mafft files.
    """
    caplog.set_level(logging.DEBUG)
    orig_mafft = subprocess.check_output("which mafft".split()).decode().strip()
    temp_mafft = orig_mafft + "-orig"
    try:
        shutil.move(orig_mafft, temp_mafft)
    except PermissionError:
        pytest.skip("No permission to change mafft bin filename")
    prt_file = os.path.join(EXPPATH, "exp_aldir-pers", "current.8.prt")
    gen_file = os.path.join(EXPPATH, "exp_aldir-pers", "current.8.gen")
    miss_file = os.path.join(EXPPATH, "exp_aldir-pers", "current.8.miss.lst")
    mafft_file = os.path.join(GENEPATH, "test_fam_align.8.aln")
    # Create empty btr file
    btr_file = os.path.join(GENEPATH, "test_fam_align_btr.8.aln")
    open(btr_file, "w").close()
    num_fam = 8
    ngenomes = 4
    logger = logging.getLogger("test_fam_align")
    assert not al.family_alignment(prt_file, gen_file, miss_file, mafft_file, btr_file, num_fam,
                                   ngenomes, logger)
    assert "Checking extractions for family 8" in caplog.text
    assert "Aligning family 8" in caplog.text
    assert ("command '>mafft --auto test/data/align/exp_files/exp_aldir-pers/current.8.prt' "
            "is not possible") in caplog.text
    # Check content of mafft and btr files
    assert not os.path.isfile(mafft_file)
    assert not os.path.isfile(btr_file)
    shutil.move(temp_mafft, orig_mafft)


def test_check_addmissing_ok():
    """
    Test that when giving a btr file with all sequences same length, and as many sequences as
    expected, it returns True
    """
    btr_file = os.path.join(EXPPATH, "exp_aldir-pers", "mafft-btr.8.aln")
    num_fam = 8
    ngenomes = 3
    logger = logging.getLogger("test_check_add_missing")
    assert al.check_add_missing(btr_file, num_fam, ngenomes, logger) is True


def test_check_addmissing_difflen():
    """
    Test that when giving a btr file with sequences with different lengths, it returns False
    """
    btr_file = os.path.join(TESTPATH, "test_alignment-mafft-diff-len.aln")
    num_fam = 8
    ngenomes = 4
    logger = logging.getLogger("test_check_add_missing")
    assert al.check_add_missing(btr_file, num_fam, ngenomes, logger) is False


def test_check_addmissing_missgenomes_prev():
    """
    Test that when giving a btr file with all sequences of same length, but not the same number
    of sequences than expected, and comes from previous run, it returns the length of alignment,
    without any error message.
    """
    btr_file = os.path.join(EXPPATH, "exp_aldir-pers", "mafft-btr.8.aln")
    num_fam = 8
    ngenomes = 4
    logger = logging.getLogger("test_check_add_missing")
    assert al.check_add_missing(btr_file, num_fam, ngenomes, logger, prev=True) == 789


def test_check_addmissing_missgenomes_noprev(caplog):
    """
    Test that when giving a btr file with all sequences of same length, but not the same number
    of sequences than expected, and does not come from previous run, it returns the length of
    alignment, with an error message.
    """
    caplog.set_level(logging.DEBUG)
    btr_file = os.path.join(EXPPATH, "exp_aldir-pers", "mafft-btr.8.aln")
    num_fam = 8
    ngenomes = 4
    logger = logging.getLogger("test_check_add_missing")
    assert al.check_add_missing(btr_file, num_fam, ngenomes, logger) == 789
    assert ("ERROR: family 8 contains 3 genomes in total instead of the 4 genomes in "
            "input.") in caplog.text


def test_add_missing_btrok():
    """
    Giving a btr_file with as many genomes as expected, and this btr file was just created (means
    that there is no missing genome in this family), it returns True
    """
    btr_file = os.path.join(EXPPATH, "exp_aldir-pers", "mafft-btr.8.aln")
    # Empty miss file (no missing genome in this family)
    miss_file = "test_add_missing"
    num_fam = 8
    ngenomes = 3
    status1 = True
    logger = logging.getLogger("test_add_missing")
    assert al.add_missing_genomes(btr_file, miss_file, num_fam, ngenomes, status1, logger)


def test_add_missing_btralreadyok(caplog):
    """
    Giving a btr_file with as many genomes as expected, and this btr file was not recreated (
    already existed from previous run), it returns "OK" with a warning message specifying that
    program will use already existing file.
    """
    caplog.set_level(logging.DEBUG)
    btr_file = os.path.join(EXPPATH, "exp_aldir-pers", "mafft-btr.8.aln")
    # Empty miss file (no missing genome in this family)
    miss_file = "test_add_missing"
    num_fam = 8
    ngenomes = 3
    status1 = "OK"
    logger = logging.getLogger("test_add_missing")
    assert al.add_missing_genomes(btr_file, miss_file, num_fam, ngenomes, status1, logger) == "OK"
    assert ("Alignment already done for family 8. The program will use it for next "
            "steps") in caplog.text


def test_add_missing_btr_difflen(caplog):
    """
    Giving a btr_file with as many genomes as expected, but different alignment lengths for the
    different sequences. It returns False, with an error message saying that there are different
    lengths in alignment.
    """
    caplog.set_level(logging.DEBUG)
    btr_file = os.path.join(TESTPATH, "test_alignment-mafft-diff-len.aln")
    # Empty miss file (no missing genome in this family)
    miss_file = "test_add_missing"
    num_fam = 8
    ngenomes = 4
    status1 = "OK"
    logger = logging.getLogger("test_add_missing")
    assert al.add_missing_genomes(btr_file, miss_file, num_fam, ngenomes, status1, logger) is False
    assert ("alignments for family 8 do not all have the same length. Lengths found "
            "are: {") in caplog.text
    assert "126" in caplog.text
    assert "154" in caplog.text
    assert "157" in caplog.text
    assert "}\n" in caplog.text


def test_add_missing_btrmiss(caplog):
    """
    Giving a btr_file with all sequences with same lengths, but 1 sequence less that the
    expected number, + miss file with the name of the missing genome.
    Returns True, after adding missing genome to btr_file (overwrites it).
    """
    caplog.set_level(logging.DEBUG)
    ref_btr_file = os.path.join(EXPPATH, "exp_aldir-pers", "mafft-btr.8.aln")
    # Copy ref_btr to another file, which will be overwritten.
    btr_file = os.path.join(GENEPATH, "test_add_missing_genomes.aln")
    shutil.copyfile(ref_btr_file, btr_file)
    miss_file = os.path.join(EXPPATH, "exp_aldir-pers", "current.8.miss.lst")
    num_fam = 8
    ngenomes = 4
    status1 = "OK"
    logger = logging.getLogger("test_add_missing")
    assert al.add_missing_genomes(btr_file, miss_file, num_fam, ngenomes, status1, logger) is True
    assert "Adding missing genomes for family 8" in caplog.text
    final_btr = os.path.join(EXPPATH, "exp_aldir-pers", "mafft-prt2nuc.8.aln")
    assert tutil.compare_file_content(btr_file, final_btr)


def test_add_missing_btrmiss2(caplog):
    """
    Giving a btr_file with all sequences with same lengths, but 2 sequences less that the
    expected number, + miss file with the name of 1 missing genome.
    Returns False, after adding 1 missing genome to btr_file. + error message saying that a
    genome is missing in btr_file.
    """
    caplog.set_level(logging.DEBUG)
    ref_btr_file = os.path.join(EXPPATH, "exp_aldir-pers", "mafft-btr.8.aln")
    # Copy ref_btr to another file, which will be overwritten.
    btr_file = os.path.join(GENEPATH, "test_add_missing_genomes.aln")
    shutil.copyfile(ref_btr_file, btr_file)
    miss_file = os.path.join(EXPPATH, "exp_aldir-pers", "current.8.miss.lst")
    num_fam = 8
    ngenomes = 5
    status1 = True
    logger = logging.getLogger("test_add_missing")
    assert al.add_missing_genomes(btr_file, miss_file, num_fam, ngenomes, status1, logger) is False
    assert "Adding missing genomes for family 8" in caplog.text
    assert ("ERROR: family 8 contains 4 genomes in total instead of the 5 genomes in "
            "input") in caplog.text
    final_btr = os.path.join(EXPPATH, "exp_aldir-pers", "mafft-prt2nuc.8.aln")
    assert tutil.compare_file_content(btr_file, final_btr)


def test_handle_family_true():
    """
    Giving an aldir with correct prt, gen and miss files, it creates mafft and btr files as
    expected, and returns True
    """
    aldir = os.path.join(GENEPATH, "aldir")
    prefix = os.path.join(aldir, "TESThandlefam")
    num_fam = 8
    ngenomes = 4
    q = multiprocessing.Manager().Queue()
    # Create aldir, and put prt, gen and miss files in it
    ref_prt = os.path.join(EXPPATH, "exp_aldir-pers", "current.8.prt")
    ref_gen = os.path.join(EXPPATH, "exp_aldir-pers", "current.8.gen")
    ref_miss = os.path.join(EXPPATH, "exp_aldir-pers", "current.8.miss.lst")
    cur_prt = os.path.join(aldir, "TESThandlefam-current.8.prt")
    cur_gen = os.path.join(aldir, "TESThandlefam-current.8.gen")
    cur_miss = os.path.join(aldir, "TESThandlefam-current.8.miss.lst")
    # Create aldir inside GENEPATH, empty directory created for this test
    os.makedirs(aldir)
    shutil.copyfile(ref_prt, cur_prt)
    shutil.copyfile(ref_gen, cur_gen)
    shutil. copyfile(ref_miss, cur_miss)
    args = (prefix, num_fam, ngenomes, q)
    assert al.handle_family(args) is True
    cur_mafft = os.path.join(aldir, "TESThandlefam-mafft-align.8.aln")
    cur_btr = os.path.join(aldir, "TESThandlefam-mafft-prt2nuc.8.aln")
    exp_mafft = os.path.join(EXPPATH, "exp_aldir-pers", "mafft-align.8.aln")
    exp_btr = os.path.join(EXPPATH, "exp_aldir-pers", "mafft-prt2nuc.8.aln")
    assert tutil.compare_order_content(cur_mafft, exp_mafft)
    assert tutil.compare_order_content(cur_btr, exp_btr)
    q.put(None)
    assert "Checking extractions for family 8" in q.get().message
    assert "Aligning family 8" in q.get().message
    assert ("mafft --auto test/data/align/generated_by_unit-tests/aldir/"
            "TESThandlefam-current.8.prt") in q.get().message
    assert "Back-translating family 8" in q.get().message
    assert "Adding missing genomes for family 8" in q.get().message
    assert not q.get()


def test_handle_family_true_nomiss():
    """
    Giving an aldir with correct prt, gen and miss files (no missing file in this family),
    it creates mafft and btr files as expected, and returns True. No need to add missing genomes
    as there are not in this family.
    """
    aldir = os.path.join(GENEPATH, "aldir")
    prefix = os.path.join(aldir, "TESThandlefam")
    num_fam = 8
    ngenomes = 3
    q = multiprocessing.Manager().Queue()
    # Create aldir, and put prt, gen and miss files in it
    ref_prt = os.path.join(EXPPATH, "exp_aldir-pers", "current.8.prt")
    ref_gen = os.path.join(EXPPATH, "exp_aldir-pers", "current.8.gen")
    cur_prt = os.path.join(aldir, "TESThandlefam-current.8.prt")
    cur_gen = os.path.join(aldir, "TESThandlefam-current.8.gen")
    cur_miss = os.path.join(aldir, "TESThandlefam-current.8.miss.lst")
    os.makedirs(aldir)
    shutil.copyfile(ref_prt, cur_prt)
    shutil.copyfile(ref_gen, cur_gen)
    open(cur_miss, "w").close()
    args = (prefix, num_fam, ngenomes, q)
    assert al.handle_family(args)
    cur_mafft = os.path.join(aldir, "TESThandlefam-mafft-align.8.aln")
    cur_btr = os.path.join(aldir, "TESThandlefam-mafft-prt2nuc.8.aln")
    exp_mafft = os.path.join(EXPPATH, "exp_aldir-pers", "mafft-align.8.aln")
    exp_btr = os.path.join(EXPPATH, "exp_aldir-pers", "mafft-btr.8.aln")
    assert tutil.compare_order_content(cur_mafft, exp_mafft)
    assert tutil.compare_order_content(cur_btr, exp_btr)
    q.put(None)
    assert "Checking extractions for family 8" in q.get().message
    assert "Aligning family 8" in q.get().message
    assert ("mafft --auto test/data/align/generated_by_unit-tests/aldir/"
            "TESThandlefam-current.8.prt") in q.get().message
    assert "Back-translating family 8" in q.get().message
    assert not q.get()


def test_handle_family_emptyaln_true():
    """
    Giving an aldir with correct prt, gen, miss files, and an empty mafft file, it creates mafft
    and btr files as expected, and returns True
    """
    aldir = os.path.join(GENEPATH, "aldir")
    prefix = os.path.join(aldir, "TESThandlefam")
    num_fam = 8
    ngenomes = 4
    q = multiprocessing.Manager().Queue()
    # Create aldir, and put prt, gen and miss files in it
    ref_prt = os.path.join(EXPPATH, "exp_aldir-pers", "current.8.prt")
    ref_gen = os.path.join(EXPPATH, "exp_aldir-pers", "current.8.gen")
    ref_miss = os.path.join(EXPPATH, "exp_aldir-pers", "current.8.miss.lst")
    cur_prt = os.path.join(aldir, "TESThandlefam-current.8.prt")
    cur_gen = os.path.join(aldir, "TESThandlefam-current.8.gen")
    cur_miss = os.path.join(aldir, "TESThandlefam-current.8.miss.lst")
    cur_mafft = os.path.join(aldir, "TESThandlefam-mafft-align.8.aln")
    # Put real gen, prt and miss files to aldir, create empty mafft file
    os.makedirs(aldir)
    shutil.copyfile(ref_prt, cur_prt)
    shutil.copyfile(ref_gen, cur_gen)
    shutil. copyfile(ref_miss, cur_miss)
    open(cur_mafft, "w").close()
    args = (prefix, num_fam, ngenomes, q)
    assert al.handle_family(args)
    cur_btr = os.path.join(aldir, "TESThandlefam-mafft-prt2nuc.8.aln")
    exp_mafft = os.path.join(EXPPATH, "exp_aldir-pers", "mafft-align.8.aln")
    exp_btr = os.path.join(EXPPATH, "exp_aldir-pers", "mafft-prt2nuc.8.aln")
    assert tutil.compare_order_content(cur_mafft, exp_mafft)
    assert tutil.compare_order_content(cur_btr, exp_btr)
    q.put(None)
    assert "Checking extractions for family 8" in q.get().message
    assert ("fam 8: Will redo alignment, because found a different number of proteins "
            "extracted in "
            "test/data/align/generated_by_unit-tests/aldir/TESThandlefam-current.8.prt (3) "
            "and proteins aligned in existing "
            "test/data/align/generated_by_unit-tests/aldir/TESThandlefam-mafft-align.8.aln "
            "(0)") in q.get().message
    assert "Aligning family 8" in q.get().message
    assert ("mafft --auto test/data/align/generated_by_unit-tests/aldir/"
            "TESThandlefam-current.8.prt") in q.get().message
    assert "Back-translating family 8" in q.get().message
    assert "Adding missing genomes for family 8" in q.get().message
    assert not q.get()


def test_handle_family_emptybtr_true():
    """
    Giving an aldir with correct prt, gen, miss and mafft files, and an empty btr files,
    it redoes back-translation and returns True
    """
    aldir = os.path.join(GENEPATH, "aldir")
    prefix = os.path.join(aldir, "TESThandlefam")
    num_fam = 8
    ngenomes = 4
    q = multiprocessing.Manager().Queue()
    # Create aldir, and put prt, gen and miss files in it
    ref_prt = os.path.join(EXPPATH, "exp_aldir-pers", "current.8.prt")
    ref_gen = os.path.join(EXPPATH, "exp_aldir-pers", "current.8.gen")
    ref_miss = os.path.join(EXPPATH, "exp_aldir-pers", "current.8.miss.lst")
    ref_mafft = os.path.join(EXPPATH, "exp_aldir-pers", "mafft-align.8.aln")
    ref_btr = os.path.join(EXPPATH, "exp_aldir-pers", "mafft-prt2nuc.8.aln")
    cur_prt = os.path.join(aldir, "TESThandlefam-current.8.prt")
    cur_gen = os.path.join(aldir, "TESThandlefam-current.8.gen")
    cur_miss = os.path.join(aldir, "TESThandlefam-current.8.miss.lst")
    cur_mafft = os.path.join(aldir, "TESThandlefam-mafft-align.8.aln")
    cur_btr = os.path.join(aldir, "TESThandlefam-mafft-prt2nuc.8.aln")
    # Put real gen, prt and miss files to aldir, create empty mafft file
    os.makedirs(aldir)
    shutil.copyfile(ref_prt, cur_prt)
    shutil.copyfile(ref_gen, cur_gen)
    shutil.copyfile(ref_miss, cur_miss)
    shutil.copyfile(ref_mafft, cur_mafft)
    open(cur_btr, "w").close()
    args = (prefix, num_fam, ngenomes, q)
    assert al.handle_family(args)
    assert tutil.compare_order_content(cur_mafft, ref_mafft)
    assert tutil.compare_order_content(cur_btr, ref_btr)
    q.put(None)
    assert "Checking extractions for family 8" in q.get().message
    assert ("fam 8: Will redo back-translation, because found a different number of proteins "
            "aligned in "
            "test/data/align/generated_by_unit-tests/aldir/TESThandlefam-mafft-align.8.aln (3) "
            "and genes back-translated in "
            "existing test/data/align/generated_by_unit-tests/"
            "aldir/TESThandlefam-mafft-prt2nuc.8.aln (0)") in q.get().message
    assert "Back-translating family 8" in q.get().message
    assert "Adding missing genomes for family 8" in q.get().message
    assert not q.get()


def test_handle_family_already_ok():
    """
    Giving an aldir with correct prt, gen, miss, mafft and btr files, it returns "OK",
    not creating any new file, but with a warning message saying that already existing files are
    used for next steps.
    """
    aldir = os.path.join(GENEPATH, "aldir")
    prefix = os.path.join(aldir, "TESThandlefam")
    num_fam = 8
    ngenomes = 4
    q = multiprocessing.Manager().Queue()
    # Create aldir, and put prt, gen, miss, mafft and btr files in it
    ref_prt = os.path.join(EXPPATH, "exp_aldir-pers", "current.8.prt")
    ref_gen = os.path.join(EXPPATH, "exp_aldir-pers", "current.8.gen")
    ref_miss = os.path.join(EXPPATH, "exp_aldir-pers", "current.8.miss.lst")
    ref_mafft = os.path.join(EXPPATH, "exp_aldir-pers", "mafft-align.8.aln")
    ref_btr = os.path.join(EXPPATH, "exp_aldir-pers", "mafft-prt2nuc.8.aln")
    cur_prt = os.path.join(aldir, "TESThandlefam-current.8.prt")
    cur_gen = os.path.join(aldir, "TESThandlefam-current.8.gen")
    cur_miss = os.path.join(aldir, "TESThandlefam-current.8.miss.lst")
    cur_mafft = os.path.join(aldir, "TESThandlefam-mafft-align.8.aln")
    cur_btr = os.path.join(aldir, "TESThandlefam-mafft-prt2nuc.8.aln")
    os.makedirs(aldir)
    shutil.copyfile(ref_prt, cur_prt)
    shutil.copyfile(ref_gen, cur_gen)
    shutil.copyfile(ref_miss, cur_miss)
    shutil.copyfile(ref_mafft, cur_mafft)
    shutil.copyfile(ref_btr, cur_btr)
    args = (prefix, num_fam, ngenomes, q)
    assert al.handle_family(args) == "OK"
    q.put(None)
    assert "Checking extractions for family 8" in q.get().message
    assert ("Alignment already done for family 8. The program will use it for next "
            "steps") in q.get().message
    assert not q.get()


def test_handle_family_wrongextract():
    """
    Giving an aldir with correct prt, gen, miss, mafft and btr files, but says that there are 5
    genomes in dataset (instead of 4), it returns False, with an error message saying that
    extractions are wrong.
    """
    aldir = os.path.join(GENEPATH, "aldir")
    prefix = os.path.join(aldir, "TESThandlefam")
    num_fam = 8
    ngenomes = 5
    q = multiprocessing.Manager().Queue()
    # Create aldir, and put prt, gen, miss, mafft and btr files in it
    ref_prt = os.path.join(EXPPATH, "exp_aldir-pers", "current.8.prt")
    ref_gen = os.path.join(EXPPATH, "exp_aldir-pers", "current.8.gen")
    ref_miss = os.path.join(EXPPATH, "exp_aldir-pers", "current.8.miss.lst")
    ref_mafft = os.path.join(EXPPATH, "exp_aldir-pers", "mafft-align.8.aln")
    ref_btr = os.path.join(EXPPATH, "exp_aldir-pers", "mafft-prt2nuc.8.aln")
    cur_prt = os.path.join(aldir, "TESThandlefam-current.8.prt")
    cur_gen = os.path.join(aldir, "TESThandlefam-current.8.gen")
    cur_miss = os.path.join(aldir, "TESThandlefam-current.8.miss.lst")
    cur_mafft = os.path.join(aldir, "TESThandlefam-mafft-align.8.aln")
    cur_btr = os.path.join(aldir, "TESThandlefam-mafft-prt2nuc.8.aln")
    os.makedirs(aldir)
    shutil.copyfile(ref_prt, cur_prt)
    shutil.copyfile(ref_gen, cur_gen)
    shutil.copyfile(ref_miss, cur_miss)
    shutil.copyfile(ref_mafft, cur_mafft)
    shutil.copyfile(ref_btr, cur_btr)
    args = (prefix, num_fam, ngenomes, q)
    assert not al.handle_family(args)
    assert not os.path.isfile(cur_mafft)
    assert not os.path.isfile(cur_btr)
    q.put(None)
    assert "Checking extractions for family 8" in q.get().message
    assert ("fam 8: wrong sum of missing genomes (1) and prt extracted (3) for 5 genomes in the "
            "dataset.") in q.get().message
    assert not q.get()


def test_align_all_true(caplog):
    """
    Giving aldir with prt, gen and miss files for families 1 and 8, as well as concat file (
    empty). It should create mafft and prt2nuc files for families 1 and 8, remove concat
    file, and return True.
    """
    caplog.set_level(logging.DEBUG)
    aldir = os.path.join(GENEPATH, "aldir")
    dname = "TESTalign-all"
    prefix = os.path.join(aldir, dname)
    all_fams = [1, 8]
    ngenomes = 4
    quiet = False
    threads = 1
    # Create aldir, and put all prt, gen and miss files for families 1 and 8
    os.makedirs(aldir)
    ref_prt8 = os.path.join(EXPPATH, "exp_aldir-pers", "current.8.prt")
    ref_gen8 = os.path.join(EXPPATH, "exp_aldir-pers", "current.8.gen")
    ref_miss8 = os.path.join(EXPPATH, "exp_aldir-pers", "current.8.miss.lst")
    ref_prt1 = os.path.join(EXPPATH, "exp_aldir", "current.1.prt")
    ref_gen1 = os.path.join(EXPPATH, "exp_aldir", "current.1.gen")
    cur_prt8 = os.path.join(aldir, dname + "-current.8.prt")
    cur_gen8 = os.path.join(aldir, dname + "-current.8.gen")
    cur_miss8 = os.path.join(aldir, dname + "-current.8.miss.lst")
    cur_prt1 = os.path.join(aldir, dname + "-current.1.prt")
    cur_gen1 = os.path.join(aldir, dname + "-current.1.gen")
    cur_miss1 = os.path.join(aldir, dname + "-current.1.miss.lst")
    refs = [ref_gen8, ref_gen1, ref_prt8, ref_prt1, ref_miss8]
    todos = [cur_gen8, cur_gen1, cur_prt8, cur_prt1, cur_miss8]
    for f1, f2 in zip(refs, todos):
        shutil.copyfile(f1, f2)
    # empty miss file for family 1
    open(cur_miss1, "w").close()
    assert os.path.isfile(cur_miss1)
    # create empty concat-file, to check that it is removed
    concat = os.path.join(aldir, dname + "-complete.cat.aln")
    open(concat, "w").close()
    assert al.align_all_families(prefix, all_fams, ngenomes, dname, quiet, threads)
    # Check output files
    out_mafft1 = os.path.join(aldir, dname + "-mafft-align.1.aln")
    out_btr1 = os.path.join(aldir, dname + "-mafft-prt2nuc.1.aln")
    out_mafft8 = os.path.join(aldir, dname + "-mafft-align.8.aln")
    out_btr8 = os.path.join(aldir, dname + "-mafft-prt2nuc.8.aln")
    exp_mafft1 = os.path.join(EXPPATH, "exp_aldir", "mafft-align.1.aln")
    exp_btr1 = os.path.join(EXPPATH, "exp_aldir", "mafft-prt2nuc.1.aln")
    exp_mafft8 = os.path.join(EXPPATH, "exp_aldir-pers", "mafft-align.8.aln")
    exp_btr8 = os.path.join(EXPPATH, "exp_aldir-pers", "mafft-prt2nuc.8.aln")
    assert tutil.compare_order_content(out_mafft1, exp_mafft1)
    assert tutil.compare_order_content(out_btr1, exp_btr1)
    assert tutil.compare_order_content(out_mafft8, exp_mafft8)
    assert tutil.compare_order_content(out_btr8, exp_btr8)
    assert not os.path.isfile(concat)
    # Check logs
    assert ("Starting alignment of all families: protein alignment, back-translation to "
            "nucleotides, and add missing genomes in the family") in caplog.text
    assert "Checking extractions for family 1" in caplog.text
    assert "Aligning family 1" in caplog.text
    assert "Back-translating family 1" in caplog.text
    assert "Checking extractions for family 8" in caplog.text
    assert "Aligning family 8" in caplog.text
    assert "Back-translating family 8" in caplog.text
    assert "Adding missing genomes for family 8" in caplog.text


def test_align_all_exists_true(caplog):
    """
    Giving aldir with prt, gen, miss, mafft and prt2nuc files for families 1 and 8, as well as
    concat file (empty). It should keep concat file, and return True.
    """
    caplog.set_level(logging.DEBUG)
    aldir = os.path.join(GENEPATH, "aldir")
    dname = "TESTalign-all"
    prefix = os.path.join(aldir, dname)
    all_fams = [1, 8]
    ngenomes = 4
    quiet = False
    threads = 2
    # Create aldir, and put all prt, gen and miss files for families 1 and 8
    os.makedirs(aldir)
    ref_prt8 = os.path.join(EXPPATH, "exp_aldir-pers", "current.8.prt")
    ref_gen8 = os.path.join(EXPPATH, "exp_aldir-pers", "current.8.gen")
    ref_miss8 = os.path.join(EXPPATH, "exp_aldir-pers", "current.8.miss.lst")
    ref_prt1 = os.path.join(EXPPATH, "exp_aldir", "current.1.prt")
    ref_gen1 = os.path.join(EXPPATH, "exp_aldir", "current.1.gen")
    exp_mafft1 = os.path.join(EXPPATH, "exp_aldir", "mafft-align.1.aln")
    exp_btr1 = os.path.join(EXPPATH, "exp_aldir", "mafft-prt2nuc.1.aln")
    exp_mafft8 = os.path.join(EXPPATH, "exp_aldir-pers", "mafft-align.8.aln")
    exp_btr8 = os.path.join(EXPPATH, "exp_aldir-pers", "mafft-prt2nuc.8.aln")
    cur_prt8 = os.path.join(aldir, dname + "-current.8.prt")
    cur_gen8 = os.path.join(aldir, dname + "-current.8.gen")
    cur_miss8 = os.path.join(aldir, dname + "-current.8.miss.lst")
    cur_prt1 = os.path.join(aldir, dname + "-current.1.prt")
    cur_gen1 = os.path.join(aldir, dname + "-current.1.gen")
    cur_miss1 = os.path.join(aldir, dname + "-current.1.miss.lst")
    out_mafft1 = os.path.join(aldir, dname + "-mafft-align.1.aln")
    out_btr1 = os.path.join(aldir, dname + "-mafft-prt2nuc.1.aln")
    out_mafft8 = os.path.join(aldir, dname + "-mafft-align.8.aln")
    out_btr8 = os.path.join(aldir, dname + "-mafft-prt2nuc.8.aln")
    refs = [ref_gen8, ref_gen1, ref_prt8, ref_prt1, ref_miss8, exp_mafft1, exp_mafft8, exp_btr1,
            exp_btr8]
    todos = [cur_gen8, cur_gen1, cur_prt8, cur_prt1, cur_miss8, out_mafft1, out_mafft8, out_btr1,
             out_btr8]
    for f1, f2 in zip(refs, todos):
        shutil.copyfile(f1, f2)
    # empty miss file for family 1
    open(cur_miss1, "w").close()
    # create empty concat-file, to check that it is removed
    concat = os.path.join(aldir, dname + "-complete.cat.aln")
    open(concat, "w").close()
    assert al.align_all_families(prefix, all_fams, ngenomes, dname, quiet, threads)
    # Check output files
    assert tutil.compare_order_content(out_mafft1, exp_mafft1)
    assert tutil.compare_order_content(out_btr1, exp_btr1)
    assert tutil.compare_order_content(out_mafft8, exp_mafft8)
    assert tutil.compare_order_content(out_btr8, exp_btr8)
    assert os.path.isfile(concat)
    with open(concat, "r") as conf:
        assert conf.readlines() == []
    # Check logs
    assert ("Starting alignment of all families: protein alignment, back-translation to "
            "nucleotides, and add missing genomes in the family") in caplog.text
    assert "Checking extractions for family 1" in caplog.text
    assert ("Alignment already done for family 1. The program will use it for next "
            "steps") in caplog.text
    assert "Checking extractions for family 8" in caplog.text
    assert ("Alignment already done for family 8. The program will use it for next "
            "steps") in caplog.text


def test_align_all_false(caplog):
    """
    Giving aldir with prt, gen, miss files for families 1 and 8, as well as
    concat file (empty). But for family 8, miss file is empty: total number of genomes does not
    correspond. It should return error message for fam 8, info logs for fam 1, remove concat file,
    and return False
    """
    caplog.set_level(logging.DEBUG)
    aldir = os.path.join(GENEPATH, "aldir")
    dname = "TESTalign-all"
    prefix = os.path.join(aldir, dname)
    all_fams = [1, 8]
    ngenomes = 4
    quiet = False
    threads = 1
    # Create aldir, and put all prt, gen and miss files for families 1 and 8
    os.makedirs(aldir)
    ref_prt8 = os.path.join(EXPPATH, "exp_aldir-pers", "current.8.prt")
    ref_gen8 = os.path.join(EXPPATH, "exp_aldir-pers", "current.8.gen")
    ref_prt1 = os.path.join(EXPPATH, "exp_aldir", "current.1.prt")
    ref_gen1 = os.path.join(EXPPATH, "exp_aldir", "current.1.gen")
    cur_prt8 = os.path.join(aldir, dname + "-current.8.prt")
    cur_gen8 = os.path.join(aldir, dname + "-current.8.gen")
    cur_miss8 = os.path.join(aldir, dname + "-current.8.miss.lst")
    cur_prt1 = os.path.join(aldir, dname + "-current.1.prt")
    cur_gen1 = os.path.join(aldir, dname + "-current.1.gen")
    cur_miss1 = os.path.join(aldir, dname + "-current.1.miss.lst")
    refs = [ref_gen8, ref_gen1, ref_prt8, ref_prt1]
    todos = [cur_gen8, cur_gen1, cur_prt8, cur_prt1]
    for f1, f2 in zip(refs, todos):
        shutil.copyfile(f1, f2)
    # empty miss file for family 1 and 8
    open(cur_miss1, "w").close()
    open(cur_miss8, "w").close()
    # create empty concat-file, to check that it is removed
    concat = os.path.join(aldir, dname + "-complete.cat.aln")
    open(concat, "w").close()
    assert not al.align_all_families(prefix, all_fams, ngenomes, dname, quiet, threads)
    # Check output files
    out_mafft1 = os.path.join(aldir, dname + "-mafft-align.1.aln")
    out_btr1 = os.path.join(aldir, dname + "-mafft-prt2nuc.1.aln")
    out_mafft8 = os.path.join(aldir, dname + "-mafft-align.8.aln")
    out_btr8 = os.path.join(aldir, dname + "-mafft-prt2nuc.8.aln")
    exp_mafft1 = os.path.join(EXPPATH, "exp_aldir", "mafft-align.1.aln")
    exp_btr1 = os.path.join(EXPPATH, "exp_aldir", "mafft-prt2nuc.1.aln")
    assert tutil.compare_order_content(out_mafft1, exp_mafft1)
    assert tutil.compare_order_content(out_btr1, exp_btr1)
    assert not os.path.isfile(out_mafft8)
    assert not os.path.isfile(out_btr8)
    assert not os.path.isfile(concat)
    # Check logs
    assert ("Starting alignment of all families: protein alignment, back-translation to "
            "nucleotides, and add missing genomes in the family") in caplog.text
    assert "Checking extractions for family 1" in caplog.text
    assert "Aligning family 1" in caplog.text
    assert "Back-translating family 1" in caplog.text
    assert "Checking extractions for family 8" in caplog.text
    assert ("fam 8: wrong sum of missing genomes (0) and prt extracted (3) for 4 genomes in the "
            "dataset.") in caplog.text



#!/usr/bin/env python3
# coding: utf-8

"""
Unit tests for the download_genomes_func submodule in prepare module
"""
import os
import logging
import shutil
import pytest
from scipy.sparse import dok_matrix

import test.test_unit.utilities_for_tests as tutil
import PanACoTA.prepare_module.filter_genomes as filterg
import PanACoTA.utils as utils

# Define variables and functions used by several tests
DATA_TEST_DIR = os.path.join("test", "data", "prepare")
GENOMES_DIR = os.path.join(DATA_TEST_DIR, "genomes", "genomes_comparison")
GENEPATH = os.path.join(DATA_TEST_DIR, "generated_by_unit-tests")
EXP_GENOMES = {
               "ACOR001.0519.fna": ["ACOR001.0519",
                                    os.path.join(GENOMES_DIR, "ACOR001.0519.fna"),
                                    os.path.join(GENOMES_DIR, "ACOR001.0519.fna"),
                                    3013644, 269, 37],
               "ACOR001.0519-bis.fna": ["ACOR001.0519-bis",
                                        os.path.join(GENOMES_DIR, "ACOR001.0519-bis.fna"),
                                        os.path.join(GENOMES_DIR, "ACOR001.0519-bis.fna"),
                                        3013644, 269, 37],
               "ACOR001.0519-almost-same.fna": ["ACOR001.0519-almost-same",
                                        os.path.join(GENOMES_DIR, "ACOR001.0519-almost-same.fna"),
                                        os.path.join(GENOMES_DIR, "ACOR001.0519-almost-same.fna"),
                                        3012665, 261, 37],
               "ACOR002.0519.fna": ["ACOR002.0519",
                                    os.path.join(GENOMES_DIR, "ACOR002.0519.fna"),
                                    os.path.join(GENOMES_DIR, "ACOR002.0519.fna"),
                                    2997537, 78, 23],
               "ACOC.1019.fna": ["ACOC.1019", os.path.join(GENOMES_DIR, "ACOC.1019.fna"),
                                  os.path.join(GENOMES_DIR, "ACOC.1019.fna"),
                                  1587120, 1, 1]
               }

LOGFILE_BASE = os.path.join(GENEPATH, "logfile_test.txt")
LEVEL = logging.DEBUG
LOGFILES = [LOGFILE_BASE + ext for ext in [".log", ".log.debug", ".log.details", ".log.err"]]


@pytest.fixture(autouse=True)
def setup_teardown_module():
    """
    Remove log files at the end of this test module

    Before each test:
    - init logger
    - create directory to put generated files

    After:
    - remove all log files
    - remove directory with generated results
    """
    # utils.init_logger(LOGFILE_BASE, LEVEL, 'test_filter', verbose=1)
    os.mkdir(GENEPATH)
    print("setup")

    yield
    for f in LOGFILES:
        if os.path.exists(f):
            os.remove(f)
    shutil.rmtree(GENEPATH)
    print("teardown")


def test_write_output_no_outdir(caplog):
    """
    Check that when outdir does not exist, program ends with error message
    """
    corresp_genomes = {"ACOR001": "ACOR001.0519.fna.gz", "ACOR002": "ACOR002.0519.fna.gz",
                       "ACOR003": "ACOR003.0519.fna.gz"}
    sorted_genomes = []
    genomes_removed = {}

    # Define output directory for generated files
    outdir = os.path.join(GENEPATH, "test_filter_write_output_no_outdir")
    gspecies = "Acetobacter_fabarum"

    # Check everything works without error
    with pytest.raises(SystemExit):
        filterg.write_outputfiles({}, sorted_genomes, genomes_removed, outdir, gspecies, 1e-4, 0.06)

    # Check outfiles do not exist
    list_file = os.path.join(outdir, "LSTINFO-Acetobacter_fabarum-filtered-0.06.txt")
    discard_file = os.path.join(outdir, "discarded-by-minhash-Acetobacter_fabarum-0.06.txt")
    assert not os.path.isfile(list_file)
    assert not os.path.isfile(discard_file)

    # Check output
    assert not os.path.isdir(outdir)

    # Check logs
    caplog.set_level(logging.DEBUG)
    assert ("The given output directory (test/data/prepare/generated_by_unit-tests/test_filter_write_output_no_outdir) "
            "does not exist. We cannot create output files there") in caplog.text


def test_write_output():
    """
    Check that the files with kept genomes and discarded genomes are created
    """
    corresp_genomes = {"ACOR001": "ACOR001.0519.fna.gz", "ACOR002": "ACOR002.0519.fna.gz",
                       "ACOR003": "ACOR003.0519.fna.gz"}
    sorted_genomes = [os.path.join(DATA_TEST_DIR, "genomes", "refseq", "bacteria", gen, gz)
                      for gen, gz in corresp_genomes.items()]
    genomes = {genome_file:["gname", "ori_name", "path_annotate", 12, 10, 1]
               for genome_file in sorted_genomes}
    toto_path = os.path.join(DATA_TEST_DIR, "genomes", "refseq", "bacteria", "toto", "toto.fna.gz")
    genomes[toto_path] = ['totoname', 'ori toto', 'path_toto', 13, 5, 6]
    genomes_removed = {"genome": ["ref", 10]}
    sorted_genomes.append(toto_path)
    # Define output directory for generated files
    outdir = os.path.join(GENEPATH, "test_filter_write_output")
    os.mkdir(outdir)
    gspecies = "Acetobacter_fabarum"
    min_dist = 1e-4
    max_dist = 0.06

    # Check everything works without error
    list_file = os.path.join(outdir, "LSTINFO-Acetobacter_fabarum-filtered-0.0001_0.06.txt")
    assert filterg.write_outputfiles(genomes, sorted_genomes,
                                     genomes_removed, outdir, gspecies,
                                     min_dist, max_dist) == list_file

    # Check outfiles exist
    discard_file = os.path.join(outdir, "discarded-by-minhash-Acetobacter_fabarum-0.0001_0.06.txt")
    assert os.path.isfile(list_file)
    assert os.path.isfile(discard_file)

    # Check content of out files
    # File containing genomes kept, that will be used for annotation
    with open(list_file) as lf:
        # Check header
        assert "to_annotate\tgsize\tnb_conts\tL90" in lf.readline()
        # Check there are 3 genomes, with expected information
        assert "path_annotate\t12\t10\t1" in lf.readline()
        assert "path_annotate\t12\t10\t1" in lf.readline()
        assert "path_annotate\t12\t10\t1" in lf.readline()
        assert "path_toto\t13\t5\t6" in lf.readline()
        assert lf.readline() == ""
        assert lf.readline() == ""

    # File with discarded genomes, and why they are discarded
    with open(discard_file) as df:
        # Check header
        assert "genome_name\tproblem_compared_with\tdist" in df.readline()
        # Check genome line
        assert "genome\tref\t10" in df.readline()
        # Check no more genomes
        assert df.readline() == ''
        assert df.readline() == ''


def test_write_output_no_discard():
    """
    Check that the files with kept genomes and discarded genomes are created, but
    file with discarded genomes is empty
    """
    corresp_genomes = {"ACOR001": "ACOR001.0519.fna.gz", "ACOR002": "ACOR002.0519.fna.gz",
                       "ACOR003": "ACOR003.0519.fna.gz"}
    sorted_genomes = [os.path.join(DATA_TEST_DIR, "genomes", "refseq", "bacteria", gen, gz)
                      for gen, gz in corresp_genomes.items()]
    genomes = {genome_file:["gname", "ori_name", "path_annotate", 12, 10, 1]
               for genome_file in sorted_genomes}
    toto_path = os.path.join(DATA_TEST_DIR, "genomes", "refseq", "bacteria", "toto", "toto.fna.gz")
    genomes[toto_path] = ['totoname', 'ori toto', 'path_toto', 13, 5, 6]
    genomes_removed = {}
    sorted_genomes.append(toto_path)

    # Define output directory for generated files
    outdir = os.path.join(GENEPATH, "test_filter_write_output_no_discard")
    os.makedirs(outdir)
    gspecies = "Acetobacter_fabarum"
    min_dist = 1e-4
    max_dist = 0.06

    # Check everything works without error
    list_file = os.path.join(outdir, "LSTINFO-Acetobacter_fabarum-filtered-0.0001_0.06.txt")
    assert filterg.write_outputfiles(genomes, sorted_genomes,
                                     genomes_removed, outdir, gspecies,
                                     min_dist, max_dist) == list_file

    # Check outfiles exist
    discard_file = os.path.join(outdir, "discarded-by-minhash-Acetobacter_fabarum-0.0001_0.06.txt")
    assert os.path.isfile(list_file)
    assert os.path.isfile(discard_file)

    # Check content of out files
    # File containing genomes kept, that will be used for annotation
    with open(list_file) as lf:
        # Check header
        assert "to_annotate\tgsize\tnb_conts\tL90" in lf.readline()
        # Check there are 3 genomes, with expected information
        assert "path_annotate\t12\t10\t1" in lf.readline()
        assert "path_annotate\t12\t10\t1" in lf.readline()
        assert "path_annotate\t12\t10\t1" in lf.readline()
        assert "path_toto\t13\t5\t6" in lf.readline()
        assert lf.readline() == ""
        assert lf.readline() == ""

    # File with discarded genomes, and why they are discarded
    with open(discard_file) as df:
        # Check header
        assert "genome_name\tproblem_compared_with\tdist" in df.readline()
        # Check no genome
        assert df.readline() == ''
        assert df.readline() == ''


def test_write_output_no_genome():
    """
    Check that the files with kept genomes and discarded genomes are created, but
    are empty because there is no genome
    """
    corresp_genomes = {"ACOR001": "ACOR001.0519.fna.gz", "ACOR002": "ACOR002.0519.fna.gz",
                       "ACOR003": "ACOR003.0519.fna.gz"}
    sorted_genomes = []
    genomes = {}
    genomes_removed = {}

    # Define output directory for generated files
    outdir = os.path.join(GENEPATH, "test_filter_write_output_no_genome")
    os.makedirs(outdir)
    gspecies = "Acetobacter_fabarum"
    min_dist = 1e-4
    max_dist = 0.06

    # Check everything works without error
    list_file = os.path.join(outdir, "LSTINFO-Acetobacter_fabarum-filtered-0.0001_0.06.txt")
    assert filterg.write_outputfiles(genomes, sorted_genomes,
                                     genomes_removed, outdir, gspecies,
                                     min_dist, max_dist) == list_file

    # Check outfiles exist
    discard_file = os.path.join(outdir, "discarded-by-minhash-Acetobacter_fabarum-0.0001_0.06.txt")
    assert os.path.isfile(list_file)
    assert os.path.isfile(discard_file)

    # Check content of out files
    # File containing genomes kept, that will be used for annotation
    with open(list_file) as lf:
        # Check header
        assert "to_annotate\tgsize\tnb_conts\tL90" in lf.readline()
        # Check there is not any genome
        assert lf.readline() == ""
        assert lf.readline() == ""

    # File with discarded genomes, and why they are discarded
    with open(discard_file) as df:
        # Check header
        assert "genome_name\tproblem_compared_with\tdist" in df.readline()
        # Check no genome
        assert df.readline() == ''
        assert df.readline() == ''


def test_sort_genomes_minhash():
    """
    Test that a given set of genomes with their proprieties is sorted as expected
    (genomes with too bad quality removed, others sorted by l90 and nbcont)
    """
    genomes = {"genome1": ["g1_name", "g1_ori", "path_genome1", 123567, 200, 101],
               "genome2": ["g2_name", "g2_ori", "path_genome2", 20000, 3, 3],
               "genome3": ["g3_name", "g3_ori", "path_genome3", 25003, 52, 50],
               "genome4": ["g4_name", "g4_ori", "path_genome4", 123456, 1023, 11],
               "genome5": ["g5_name", "g5_ori", "path_genome5", 22012, 20, 10],
               "genome6": ["g6_name", "g6_ori", "path_genome6", 1500, 3, 2]
              }
    max_l90 = 100
    max_cont = 1000
    sorted_genomes = filterg.sort_genomes_minhash(genomes, max_l90, max_cont)
    assert sorted_genomes == ["genome6", "genome2", "genome5", "genome3"]


def test_sketch_all(caplog):
    """
    Test that all genomes are sketch, in the provided order
    """
    utils.init_logger(LOGFILE_BASE, LEVEL, 'test_filter', verbose=1)
    caplog.set_level(logging.DEBUG)
    # We give 5 genomes
    genomes = {"genome1": ["g1_name", "g1_ori", os.path.join(GENOMES_DIR, "ACOR001.0519.fna"),
                           123567, 200, 101],
               "genome1bis": ["g1bis_name", "g1bis_ori",
                              os.path.join(GENOMES_DIR, "ACOR001.0519-bis.fna"),
                              251500, 200, 101],
               "genome1diff": ["g1diff_name", "g1diff_ori",
                              os.path.join(GENOMES_DIR, "ACOR001.0519-almost-same.fna"),
                              1500, 3, 2],
               "genome2": ["g2_name", "g2_ori", os.path.join(GENOMES_DIR, "ACOR002.0519.fna"),
                           20000, 3, 1],
               "genome3": ["g3_name", "g3_ori", os.path.join(GENOMES_DIR, "ACOC.1019.fna"), 25003, 52, 50]
               }
    sorted_genomes = ["genome2", "genome1diff", "genome3", "genome1", "genome1bis"]
    outdir = os.path.join(GENEPATH, "test_sketch_all")
    os.makedirs(outdir)
    list_reps = os.path.join(outdir, "test_list_reps.txt")
    out_msh = os.path.join(outdir, "out_mash.msh")
    mash_log = os.path.join(outdir, "mash_sketch.log")
    threads = 1
    filterg.sketch_all(genomes, sorted_genomes, outdir, list_reps, out_msh, mash_log, threads)

    # Check that expected output files were created
    assert os.path.isfile(list_reps)
    assert os.path.isfile(mash_log)
    assert os.path.isfile(out_msh)

    # Check content of list file
    expected_lines = [os.path.join(GENOMES_DIR, "ACOR002.0519.fna"),
                      os.path.join(GENOMES_DIR, "ACOR001.0519-almost-same.fna"),
                      os.path.join(GENOMES_DIR, "ACOC.1019.fna"),
                      os.path.join(GENOMES_DIR, "ACOR001.0519.fna"),
                      os.path.join(GENOMES_DIR, "ACOR001.0519-bis.fna")]
    assert tutil.compare_file_to_list(list_reps, expected_lines)

    with open(mash_log, "r") as ml:
        assert ml.readline().strip() == f"Sketching {expected_lines[0]}..."
        assert ml.readline().strip() == f"Sketching {expected_lines[1]}..."
        assert ml.readline().strip() == f"Sketching {expected_lines[2]}..."
        assert ml.readline().strip() == f"Sketching {expected_lines[3]}..."
        assert ml.readline().strip() == f"Sketching {expected_lines[4]}..."
        assert ml.readline().strip() == f"Writing to {out_msh}..."

    assert "Sketching all genomes..." in caplog.text
    assert ("mash sketch -o test/data/prepare/generated_by_unit-tests/test_sketch_all/out_mash.msh -p 1 -l "
            "test/data/prepare/generated_by_unit-tests/test_sketch_all/test_list_reps.txt -s 1e4") in caplog.text
    assert caplog.records[0].levelname == "INFO"
    assert caplog.records[1].levelname == "DETAIL"


def test_sketch_all_noout(caplog):
    """
    Test that, when given output directory does not exist, it ends program with error message
    """
    genomes = {"genome1": ["g1_name", "g1_ori", os.path.join(GENOMES_DIR, "ACOR001.0519.fna"),
                           123567, 200, 101],
               "genome2": ["g2_name", "g2_ori", os.path.join(GENOMES_DIR, "ACOR002.0519.fna"),
                           20000, 3, 1],
               "genome3": ["g3_name", "g3_ori", os.path.join(GENOMES_DIR, "ACOR003.0519.fna"), 25003, 52, 50]
               }
    sorted_genomes = ["genome2", "genome3", "genome1"]
    outdir = os.path.join(GENEPATH, "test_sketch_all_noout")
    list_reps = os.path.join(outdir, "test_list_reps.txt")
    out_msh = os.path.join(outdir, "out_mash.msh")
    mash_log = os.path.join(outdir, "mash_sketch.log")
    threads = 1
    # Check everything works without error
    with pytest.raises(SystemExit):
        filterg.sketch_all(genomes, sorted_genomes, outdir, list_reps, out_msh, mash_log, threads)

    # Check log
    caplog.set_level(logging.DEBUG)
    assert ("Your output directory 'test/data/prepare/generated_by_unit-tests/test_sketch_all_noout' "
            "does not exist") in caplog.text


def test_sketch_all_mash_exists(caplog):
    """
    Test that, when mash sketch file already exists, it does nothing: message to say
    that the file already exists, and returns 0
    """
    genomes = {"genome1": ["g1_name", "g1_ori", os.path.join(GENOMES_DIR, "ACOR001.0519.fna"),
                           123567, 200, 101],
               "genome2": ["g2_name", "g2_ori", os.path.join(GENOMES_DIR, "ACOR002.0519.fna"),
                           20000, 3, 1],
               "genome3": ["g3_name", "g3_ori", os.path.join(GENOMES_DIR, "ACOR003.0519.fna"), 25003, 52, 50]
               }
    sorted_genomes = ["genome2", "genome3", "genome1"]
    outdir = os.path.join(GENEPATH, "test_sketch_all_mash_exists")
    os.makedirs(outdir)
    list_reps = os.path.join(outdir, "test_list_reps.txt")
    out_msh = os.path.join(outdir, "mash_exists")
    # Create empty msh file
    open(out_msh + ".msh", "w").close()
    mash_log = os.path.join(outdir, "mash_sketch.log")
    threads = 1
    # Check that out_msh already exists before running mash sketch
    assert os.path.isfile(out_msh + ".msh")
    filterg.sketch_all(genomes, sorted_genomes, outdir, list_reps, out_msh, mash_log, threads)

    # Check that mash file still exists, but no other outfile was created
    assert not os.path.isfile(list_reps)
    assert not os.path.isfile(mash_log)
    assert os.path.isfile(out_msh + ".msh")

    # Check log
    caplog.set_level(logging.DEBUG)
    assert ("Mash sketch file test/data/prepare/generated_by_unit-tests/test_sketch_all_mash_exists/mash_exists.msh "
            "already exists. PanACoTA will use it for next step.") in caplog.text


def test_sketch_all_error_mash(caplog):
    """
    Test that, when mash has a problem, PanACoTA exits with an error message
    """
    genomes = {"genome1": ["g1_name", "g1_ori", "genome1", 123567, 200, 101],
               "genome2": ["g2_name", "g2_ori", "genome2", 20000, 3, 1],
               "genome3": ["g3_name", "g3_ori", os.path.join(GENOMES_DIR, "ACOR003.0519.fna"), 25003, 52, 50]
               }
    sorted_genomes = ["genome2", "genome3", "genome1"]
    outdir = os.path.join(GENEPATH, "test_sketch_all_mash_error")
    os.makedirs(outdir)
    list_reps = os.path.join(DATA_TEST_DIR, "test_files", "test_list_to_sketch.txt")
    out_msh = os.path.join(outdir, "out_mash.msh")
    mash_log = os.path.join(outdir, "mash_sketch.log")
    threads = 1

    # Test that it exists with sysExit error
    with pytest.raises(SystemExit):
        filterg.sketch_all(genomes, sorted_genomes, outdir, list_reps, out_msh, mash_log, threads)

    # Check that expected output files were created
    assert os.path.isfile(list_reps)
    assert os.path.isfile(mash_log)
    assert not os.path.isfile(out_msh)

    # Check log
    caplog.set_level(logging.DEBUG)
    assert ("Error while trying to sketch 3 genomes to combined archive. Maybe some genome "
            "sequences in 'tmp_files' are missing! Check logfile: test/data/prepare/generated_by_unit-tests/"
            "test_sketch_all_mash_error/mash_sketch.log") in caplog.text


def test_compare_all(caplog):
    """
    Check that comparison of all sketched sequences is as expected (output matrix is as expected)
    """
    out_msh = os.path.join(DATA_TEST_DIR, "test_files", "test_mash_output")
    matrix = os.path.join(GENEPATH, "matrix_from_test_compare_all.txt")
    mash_log = os.path.join(GENEPATH, "mashlog_from_test_compare_all.log")
    threads = 1

    # Check msh file exists
    assert os.path.isfile(out_msh + ".msh")

    filterg.compare_all(out_msh, matrix, "matrix", mash_log, threads)

    # Check output files are created
    assert os.path.isfile(matrix)
    assert os.path.isfile(mash_log)

    # Check content of matrix file
    expect_matrix = os.path.join(DATA_TEST_DIR, "test_files", "test_matrix_mash.txt")
    assert tutil.compare_file_content(matrix, expect_matrix)


def test_compare_all_matrix_exists(caplog):
    """
    Check that matrix file already exists, it returns 0 with a warning message sayng that the
    existing file will be used.
    """
    out_msh = os.path.join(DATA_TEST_DIR, "test_files", "test_mash_output")
    matrix = os.path.join(DATA_TEST_DIR, "test_files", "test_matrix_mash.txt")
    mash_log = os.path.join(GENEPATH, "mashlog_from_test_compare_all.log")
    threads = 1

    filterg.compare_all(out_msh, matrix, "matrix", mash_log, threads)

    # Check log
    caplog.set_level(logging.DEBUG)
    assert ("Matrix file test/data/prepare/test_files/test_matrix_mash.txt already exists. "
            "The program will use this distance matrix to filter all genomes according to "
            "their distances") in caplog.text


def test_compare_all_error_mash(caplog):
    """
    Check that when mash has a problem, it gives an error message and closes the program
    """
    # mash file does not exist
    out_msh = os.path.join(GENEPATH, "mash.msh")
    matrix = os.path.join(GENEPATH, "matrix.txt")
    mash_log = os.path.join(GENEPATH, "mashlog_from_test_compare_all-error-mash.log")
    threads = 1

    # Test that it exists with sysExit error
    with pytest.raises(SystemExit):
        filterg.compare_all(out_msh, matrix, "matrix", mash_log, threads)

    # Check log
    caplog.set_level(logging.DEBUG)
    assert ("Error while trying to estimate pairwise distances between all genomes. "
            "See test/data/prepare/generated_by_unit-tests/mashlog_from_test_compare_all-error-mash.log") in caplog.text


def test_read_matrix():
    """
    Test that the matrix txt file is converted to a scipy matrix as expected
    """
    genomes = {"genome1": ["g1_name", "g1_ori", os.path.join(GENOMES_DIR, "ACOR001.0519.fna"),
                           123567, 200, 101],
               "genome1bis": ["g1bis_name", "g1bis_ori",
                              os.path.join(GENOMES_DIR, "ACOR001.0519-bis.fna"),
                              251500, 200, 101],
               "genome1diff": ["g1diff_name", "g1diff_ori",
                              os.path.join(GENOMES_DIR, "ACOR001.0519-almost-same.fna"),
                              1500, 3, 2],
               "genome2": ["g2_name", "g2_ori", os.path.join(GENOMES_DIR, "ACOR002.0519.fna"),
                           20000, 3, 1],
               "genome3": ["g3_name", "g3_ori", os.path.join(GENOMES_DIR, "ACOC.1019.fna"), 25003, 52, 50]
               }
    sorted_genomes = ["genome2", "genome1diff", "genome3", "genome1", "genome1bis"]
    matrix_file = os.path.join(DATA_TEST_DIR, "test_files", "test_matrix_mash.txt")

    # Run matrix reading
    out_mat = filterg.read_matrix(genomes, sorted_genomes, matrix_file)

    exp_mat = dok_matrix((5, 5), dtype=float)
    exp_mat[0, 0] = 0  # genome2 vs genome2
    exp_mat[0, 1] = 0.000167546  # genome2 vs genome1diff
    exp_mat[0, 2] = 0.295981  # genome2 vs genome3
    exp_mat[0, 3] = 0.000143503  # genome2 vs genome1
    exp_mat[0, 4] = 0.000143503  # genome2 vs genome1bis
    exp_mat[1, 1] = 0 # genome1diff vs genome1diff
    exp_mat[1, 2] = 0.295981 # genome1diff vs genome3
    exp_mat[1, 3] = 2.38274e-05  # genome1diff vs genome1
    exp_mat[1, 4] = 2.38274e-05  # genome1diff vs genome1bis
    exp_mat[2, 3] = 0.295981  # genome3 vs genome1
    exp_mat[2, 4] = 0.295981  # genome3 vs genome1bis
    exp_mat[3, 4] = 0  # genome1 vs genome1bis

    out = out_mat.toarray()
    exp = exp_mat.toarray()
    import numpy as np
    assert  np.array_equal(out, exp)


def test_read_matrix_nofile(caplog):
    """
    Test that when the given matrix file does not exist, it exits with error message
    """
    genomes = {"genome1": ["g1_name", "g1_ori", os.path.join(GENOMES_DIR, "ACOR001.0519.fna"),
                           123567, 200, 101],
               "genome1bis": ["g1bis_name", "g1bis_ori",
                              os.path.join(GENOMES_DIR, "ACOR001.0519-bis.fna"),
                              251500, 200, 101],
               "genome1diff": ["g1diff_name", "g1diff_ori",
                              os.path.join(GENOMES_DIR, "ACOR001.0519-almost-same.fna"),
                              1500, 3, 2],
               "genome2": ["g2_name", "g2_ori", os.path.join(GENOMES_DIR, "ACOR002.0519.fna"),
                           20000, 3, 1],
               "genome3": ["g3_name", "g3_ori", os.path.join(GENOMES_DIR, "ACOC.1019.fna"), 25003, 52, 50]
               }
    sorted_genomes = ["genome2", "genome1diff", "genome3", "genome1", "genome1bis"]
    matrix_file = os.path.join(DATA_TEST_DIR, "mymatrix.txt")

    # Test that it exists with sysExit error
    with pytest.raises(SystemExit):
        out_mat = filterg.read_matrix(genomes, sorted_genomes, matrix_file)

    # Check logs
    caplog.set_level(logging.DEBUG)
    assert("Matrix file test/data/prepare/mymatrix.txt does not exist. We cannot "
           "read it and do the next steps. Program ending.") in caplog.text


def test_mash_step_2steps():
    """
    Test that when comparing a given reference genome to all others kept until now:
    - it updates 'genomes_removed' (genomes removed as they are not between min_dist
    and max_dist compared to the given reference genome.
    - it updates 'to_try' object, removing the given reference genome, as well as discarded
    genomes
    """
    to_try = ["genome1bis", "genome1", "genome3", "genome1diff", "genome2"]
    corresp = {"genome1":3, "genome1bis":4, "genome1diff":1, "genome2":0, "genome3":2}

    # Create matrix to read
    exp_mat = dok_matrix((5, 5), dtype=float)
    exp_mat[0, 0] = 0  # genome2 vs genome2
    exp_mat[0, 1] = 0.000167546  # genome2 vs genome1diff
    exp_mat[0, 2] = 0.295981  # genome2 vs genome3
    exp_mat[0, 3] = 0.000143503  # genome2 vs genome1
    exp_mat[0, 4] = 0.000143503  # genome2 vs genome1bis
    exp_mat[1, 1] = 0 # genome1diff vs genome1diff
    exp_mat[1, 2] = 0.295981 # genome1diff vs genome3
    exp_mat[1, 3] = 2.38274e-05  # genome1diff vs genome1
    exp_mat[1, 4] = 2.38274e-05  # genome1diff vs genome1bis
    exp_mat[2, 3] = 0.295981  # genome3 vs genome1
    exp_mat[2, 4] = 0.295981  # genome3 vs genome1bis
    exp_mat[3, 4] = 0  # genome1 vs genome1bis

    genomes_removed = {"toto": ["titi", 0]}
    min_dist = 1e-4
    max_dist = 0.06

    # First step: compare genome2 to all other genomes
    filterg.mash_step(to_try, corresp, exp_mat, genomes_removed, min_dist, max_dist)

    exp_removed = {"toto": ["titi", 0],
                   "genome3": ["genome2", 0.295981]}

    exp_to_try = ["genome1bis", "genome1", "genome1diff"]
    assert genomes_removed == exp_removed
    assert to_try == exp_to_try

    # Second step: compare genome1diff to all other genomes
    filterg.mash_step(exp_to_try, corresp, exp_mat, exp_removed, min_dist, max_dist)

    out_removed = {"toto": ["titi", 0],
                   "genome1": ["genome1diff", 2.38274e-05],
                   "genome1bis": ["genome1diff", 2.38274e-05],
                   "genome3": ["genome2", 0.295981]
                   }

    out_to_try = []

    assert exp_removed == out_removed
    assert exp_to_try == out_to_try


def test_mash_step_wrong_mat(caplog):
    """
    Test that when comparing a given reference genome to all others kept until now:
    - it updates 'genomes_removed' (genomes removed as they are not between min_dist
    and max_dist compared to the given reference genome.
    - it updates 'to_try' object, removing the given reference genome, as well as discarded
    genomes
    BUT, as the matrix is not triangular, prints a warning
    """
    to_try = ["genome1bis", "genome1", "genome3", "genome2", "genome1diff", "genome2"]
    corresp = {"genome1":3, "genome1bis":4, "genome1diff":1, "genome2":0, "genome3":2}

    # Create matrix to read
    exp_mat = dok_matrix((5, 5), dtype=float)
    exp_mat[0, 0] = 0  # genome2 vs genome2
    exp_mat[0, 1] = 0.000167546  # genome2 vs genome1diff
    exp_mat[0, 2] = 0.295981  # genome2 vs genome3
    exp_mat[0, 3] = 0.000143503  # genome2 vs genome1
    exp_mat[0, 4] = 0.000143503  # genome2 vs genome1bis
    exp_mat[1, 1] = 0 # genome1diff vs genome1diff
    exp_mat[1, 2] = 0.295981 # genome1diff vs genome3
    exp_mat[1, 3] = 2.38274e-05  # genome1diff vs genome1
    exp_mat[1, 4] = 2.38274e-05  # genome1diff vs genome1bis
    exp_mat[2, 3] = 0.295981  # genome3 vs genome1
    exp_mat[2, 4] = 0.295981  # genome3 vs genome1bis
    exp_mat[3, 4] = 0  # genome1 vs genome1bis

    genomes_removed = {"toto": ["titi", 0]}
    min_dist = 1e-4
    max_dist = 0.06

    # Compare genome2 to all other genomes
    filterg.mash_step(to_try, corresp, exp_mat, genomes_removed, min_dist, max_dist)

    exp_removed = {"toto": ["titi", 0],
                   "genome3": ["genome2", 0.295981],
                   "genome2": ["genome2", 0]
                   }
    exp_to_try = ["genome1bis", "genome1", "genome1diff"]
    assert genomes_removed == exp_removed
    assert to_try == exp_to_try

    # Check logs
    caplog.set_level(logging.DEBUG)
    assert "Should never happen as mat_sp is a triangle matrix!" in caplog.text


def test_check_quality():
    """
    quality control of all genomes in the database
    """
    species_linked = "my-test-genomes"
    db_path = os.path.join(DATA_TEST_DIR, "genomes", "genomes_comparison")
    tmp_dir = os.path.join(GENEPATH, "tmp_dir_check_quality")
    os.mkdir(tmp_dir)
    max_l90 = 100
    max_cont = 100
    cutn = 0

    genomes = filterg.check_quality(species_linked, db_path, tmp_dir, max_l90, max_cont, cutn)

    assert genomes == EXP_GENOMES


def test_check_quality_no_dbdir(caplog):
    """
    quality control of all genomes in the database when given db folder does not exist:
    ends with error message
    """
    species_linked = "my-test-genomes"
    db_path = "dbpath_for_test"
    tmp_dir = "tmp_dir_check_quality"
    max_l90 = 100
    max_cont = 100
    cutn = 5

    # Test that it exists with sysExit error
    with pytest.raises(SystemExit):
        genomes = filterg.check_quality(species_linked, db_path, tmp_dir, max_l90, max_cont, cutn)

    # Check logs
    caplog.set_level(logging.DEBUG)
    assert "dbpath_for_test does not exist" in caplog.text


def test_check_quality_no_tmpdir(caplog):
    """
    quality control of all genomes in the database when given tmp folder does not exist:
    ends with error message
    """
    species_linked = "my-test-genomes"
    db_path = os.path.join(DATA_TEST_DIR, "genomes", "genomes_comparison")
    tmp_dir = "tmp_dir_check_quality"
    max_l90 = 100
    max_cont = 100
    cutn = 5

    # Test that it exists with sysExit error
    with pytest.raises(SystemExit):
        genomes = filterg.check_quality(species_linked, db_path, tmp_dir, max_l90, max_cont, cutn)

    # Check logs
    caplog.set_level(logging.DEBUG)
    assert "tmp_dir_check_quality does not exist" in caplog.text


def test_check_quality_same_contname(caplog):
    """
    quality control of all genomes in the database when one genome contains 2 contigs
    with the same name: this genome is discarded + warning message
    """
    caplog.set_level(logging.DEBUG)
    species_linked = "my-test-genomes"
    db_path = os.path.join(DATA_TEST_DIR, "genomes", "genomes_comparison")
    tmp_dir = os.path.join(GENEPATH, "tmp_dir_check_quality")
    os.makedirs(tmp_dir)
    # Copy all genomes to new database
    db_path_used = os.path.join(GENEPATH, "database")
    shutil.copytree(db_path, db_path_used)
    # Add another genome, with 2 identical contig names
    new_genome = os.path.join(db_path_used, "genome_duplicate.fst")
    with open(new_genome, "w") as ng:
        ng.write(">my_contig\nAACTATATAGGAAGACACACAATTAAGGGACAGG\n")
        ng.write(">my_contig2\nCCNNCCGATTCGAGCACACAATTAAGGGACAGG\n")
        ng.write(">my_contig\nCCANNCCATCTCTCTTATCTCTCTAANNNNCTCTANCCNNNNNNCCATTCA\n")
    max_l90 = 100
    max_cont = 100
    cutn = 0

    # Get genomes information
    genomes = filterg.check_quality(species_linked, db_path_used, tmp_dir, max_l90, max_cont, cutn)

    # Even if there was 1 more genome in the database, output list of genomes is the same,
    # as the supplementary genome contains identical contig names
    exp_genomes = {
               "ACOR001.0519.fna": ["ACOR001.0519",
                                    os.path.join(db_path_used, "ACOR001.0519.fna"),
                                    os.path.join(db_path_used, "ACOR001.0519.fna"),
                                    3013644, 269, 37],
               "ACOR001.0519-bis.fna": ["ACOR001.0519-bis",
                                        os.path.join(db_path_used, "ACOR001.0519-bis.fna"),
                                        os.path.join(db_path_used, "ACOR001.0519-bis.fna"),
                                        3013644, 269, 37],
               "ACOR001.0519-almost-same.fna": ["ACOR001.0519-almost-same",
                                        os.path.join(db_path_used, "ACOR001.0519-almost-same.fna"),
                                        os.path.join(db_path_used, "ACOR001.0519-almost-same.fna"),
                                        3012665, 261, 37],
               "ACOR002.0519.fna": ["ACOR002.0519",
                                    os.path.join(db_path_used, "ACOR002.0519.fna"),
                                    os.path.join(db_path_used, "ACOR002.0519.fna"),
                                    2997537, 78, 23],
               "ACOC.1019.fna": ["ACOC.1019", os.path.join(db_path_used, "ACOC.1019.fna"),
                                  os.path.join(db_path_used, "ACOC.1019.fna"),
                                  1587120, 1, 1]
               }

    assert genomes == exp_genomes

    # Check logs
    assert ("Total number of genomes for my-test-genomes: 6") in caplog.text
    assert ("In genome genome_duplicate.fst, '>my_contig' contig name is used for several "
            "contigs. Please put different names for each contig. This genome will be "
            "ignored.") in caplog.text


def test_check_quality_no_genome(caplog):
    """
    quality control of all genomes in the database where there is no genome to annotate:
    ends with error message
    """
    species_linked = "my-test-genomes"
    db_path = os.path.join(GENEPATH, "dbpath_for_test")
    os.mkdir(db_path)
    tmp_dir = os.path.join(GENEPATH, "tmp_dir_check_quality")
    os.mkdir(tmp_dir)
    max_l90 = 100
    max_cont = 100
    cutn = 5

    # Test that it exists with sysExit error
    with pytest.raises(SystemExit):
        genomes = filterg.check_quality(species_linked, db_path, tmp_dir, max_l90, max_cont, cutn)

    # Check logs
    caplog.set_level(logging.DEBUG)
    assert "There is no genome in test/data/prepare/generated_by_unit-tests/dbpath_for_test." in caplog.text


def test_iterative_mash(caplog):
    """
    Test that when we give all genomes, sorted, it compares them, and it saves the comparison
    matrix in a npz file, and return genomes removed
    """
    # Parameters for iterative_mash function
    sorted_genomes = ["ACOR002.0519.fna", "ACOR001.0519-almost-same.fna",
                      "ACOC.1019.fna", "ACOR001.0519.fna", "ACOR001.0519-bis.fna"]
    outdir = os.path.join(GENEPATH, "res_test_iterative_mash")
    species_linked = "my-test-species"
    min_dist = 1e-4
    max_dist = 0.06
    threads = 1
    quiet = False

    # Expected created files
    mash_dir = os.path.join(outdir, "mash_files")
    txt_matrix = os.path.join(mash_dir, "matrix-all-genomes-my-test-species.txt")
    npz_mat = os.path.join(mash_dir, "matrix-all-genomes-my-test-species.npz")

    removed = filterg.iterative_mash(sorted_genomes, EXP_GENOMES, outdir,
                                     species_linked, min_dist, max_dist, threads, quiet)

    # At least, ACOC.1019.fna should be remove by ACOR002.0519.fna, as we made them
    # completely different. For the others, depends on mash version...
    # mash1.1 get correct difference between ACOR002 and ACOR001
    # but mash2.2 considers them as too close
    assert "ACOC.1019.fna" in removed.keys()
    assert removed["ACOC.1019.fna"][0] == "ACOR002.0519.fna"

    # Check that txt and npz matrix were created
    # We cannot check their content as distances depend on mash version...
    assert os.path.isfile(npz_mat)
    assert os.path.isfile(txt_matrix)


def test_iterative_mash_npz_exists():
    """
    Test that when we give all genomes, sorted, and that the mash matrix was already calculated,
    and is now stored in an npz file, it does not re-calculate it, but returns directly removed
    genomes.
    """
    sorted_genomes = ["ACOR002.0519.fna", "ACOR001.0519-almost-same.fna",
                      "ACOC.1019.fna", "ACOR001.0519.fna", "ACOR001.0519-bis.fna"]
    # Create output dir where all mash result files will be stored
    outdir = os.path.join(GENEPATH, "res_test_iterative_mash_npz_exists")
    mash_dir = os.path.join(outdir, "mash_files")
    os.makedirs(mash_dir)
    # Copy existing npz matrix to mash output folder
    npz_matrix_model = os.path.join(DATA_TEST_DIR, "test_files",
                                    "test_npz_matrix_mash.npz")
    npz_matrix_out = os.path.join(mash_dir, "matrix-all-genomes-my-test-species.npz")
    shutil.copy(npz_matrix_model, npz_matrix_out)
    txt_matrix_out = os.path.join(mash_dir, "matrix-all-genomes-my-test-species.txt")
    species_linked = "my-test-species"
    min_dist = 1e-4
    max_dist = 0.06
    threads = 1
    quiet = False

    # Check npz matrix already exists, and txt matrix does not
    assert os.path.isfile(npz_matrix_out)
    assert not os.path.isfile(txt_matrix_out)

    removed = filterg.iterative_mash(sorted_genomes, EXP_GENOMES, outdir,
                                     species_linked, min_dist, max_dist, threads, quiet)

    # Compare output dict
    exp_removed = {"ACOC.1019.fna": ["ACOR002.0519.fna", 0.295981],
                   "ACOR001.0519-bis.fna": ["ACOR001.0519-almost-same.fna", 2.38274e-05],
                   "ACOR001.0519.fna": ["ACOR001.0519-almost-same.fna", 2.38274e-05]}

    assert removed == exp_removed

    # Check npz matrix exists (and is the same as the given file), and txt matrix was not created,
    # as we used the npz file
    assert os.path.isfile(npz_matrix_out)
    assert tutil.compare_files_bin(npz_matrix_out, npz_matrix_model)
    assert not os.path.isfile(txt_matrix_out)



#!/usr/bin/env python3

"""
Functional tests for PanACoTA corepers subcommand
"""
import glob
import os
import subprocess
import pytest
import shutil
import logging

import PanACoTA.subcommands.corepers as corepers
import test.test_unit.utilities_for_tests as tutil


PERS_PATH = os.path.join("test", "data", "persgenome")
TEST_PATH = os.path.join(PERS_PATH, "test_files")
EXP_PATH = os.path.join(PERS_PATH, "exp_files")
GENEPATH = os.path.join(PERS_PATH, "generated_by_func-tests")
OPAN = os.path.join(TEST_PATH, "test_pan-for-corepers.txt")
UPAN = os.path.join(GENEPATH, "pangenome.lst")

@pytest.fixture(autouse=True)
def setup_teardown_module():
    """
    Remove log files at the end of this test module

    Before each test:
    - init logger
    - create directory to put generated files

    After:
    - remove all log files
    - remove directory with generated results
    """
    if not os.path.isdir(GENEPATH):
        os.mkdir(GENEPATH)
    shutil.copyfile(OPAN, UPAN)
    print("setup")

    yield
    shutil.rmtree(GENEPATH, ignore_errors=True)
    print("teardown")


def test_main_default(capsys):
    """
    Test that with default parameters, it creates the expected core genome.
    """
    tol = 1
    multi = False
    mixed = False
    cmd = "cmd"
    out_pers = os.path.join(GENEPATH, "PersGenome_pangenome.lst_1.lst")
    assert corepers.main(cmd, UPAN, tol, multi, mixed, GENEPATH) == out_pers
    # Check creation of binary file for pangenome, and remove it
    assert os.path.isfile(UPAN + ".bin")
    # Check presence of persistent genome, and its content, and remove it
    exp_pers = os.path.join(EXP_PATH, "exp_coregenome.txt")
    assert os.path.isfile(out_pers)
    assert tutil.compare_order_content(out_pers, exp_pers)
    # Check presence of log files and remove them
    logfile = os.path.join(GENEPATH, "PanACoTA-corepers.log")
    assert os.path.isfile(logfile)
    assert os.path.isfile(logfile + ".err")
    # Check log messages
    out, err = capsys.readouterr()
    assert "Will generate a CoreGenome." in out
    assert "Saving all information to a binary file for later use" in out
    assert "Generating Persistent genome of a dataset containing 4 genomes" in out
    assert ("The core genome contains 2 families, each one having exactly "
            "4 members, from the 4 different genomes.") in out


def test_main_pers(capsys):
    """
    Test that with default parameters, it creates the expected core genome.
    """
    tol = 0.99
    multi = False
    mixed = False
    cmd = "cmd"
    out_pers = os.path.join(GENEPATH, "PersGenome_pangenome.lst_0.99.lst")
    assert corepers.main(cmd, UPAN, tol, multi, mixed, GENEPATH) == out_pers
    # Check creation of binary file for pangenome, and remove it
    assert os.path.isfile(UPAN + ".bin")
    # Check presence of persistent genome, and its content, and remove it
    exp_pers = os.path.join(EXP_PATH, "exp_coregenome.txt")
    assert os.path.isfile(out_pers)
    assert tutil.compare_order_content(out_pers, exp_pers)
    # Check presence of log files and remove them
    logfile = os.path.join(GENEPATH, "PanACoTA-corepers.log")
    assert os.path.isfile(logfile)
    assert os.path.isfile(logfile + ".err")
    # Check log messages
    out, err = capsys.readouterr()
    assert ("Will generate a Persistent genome with member(s) in at least 99.0% "
            "of all genomes in each family.") in out
    assert ("To be considered as persistent, a family "
            "must contain exactly 1 member in at least 99.0% of all genomes. "
            "The other genomes are absent from the family.") in out
    assert "Saving all information to a binary file for later use" in out
    assert "Generating Persistent genome of a dataset containing 4 genomes" in out
    assert ("The persistent genome contains 2 families, each one having exactly "
            "1 member from at least 99.0% of the 4 different genomes "
            "(that is 4 genomes). The other genomes are absent from the family.") in out


def test_main_pers_floor_verbose2(capsys):
    """
    Test that with default parameters, it creates the expected core genome.
    """
    tol = 0.99
    multi = False
    mixed = False
    floor = True
    cmd = "cmd"
    floor = True
    out_pers = os.path.join(GENEPATH, "PersGenome_pangenome.lst_F0.99.lst")
    assert corepers.main(cmd, UPAN, tol, multi, mixed, GENEPATH,
                         floor=floor, verbose=2) == out_pers
    # Check creation of binary file for pangenome, and remove it
    assert os.path.isfile(UPAN + ".bin")
    # Check presence of persistent genome, and its content, and remove it
    exp_pers = os.path.join(EXP_PATH, "exp_pers-floor-strict.txt")
    assert os.path.isfile(out_pers)
    assert tutil.compare_order_content(out_pers, exp_pers)
    # Check presence of log files and remove them
    logfile = os.path.join(GENEPATH, "PanACoTA-corepers.log")
    assert os.path.isfile(logfile)
    assert os.path.isfile(logfile + ".err")
    assert os.path.isfile(logfile + ".details")
    # Check log messages
    out, err = capsys.readouterr()
    assert ("Will generate a Persistent genome with member(s) in at least 99.0% "
            "of all genomes in each family.") in out
    assert ("To be considered as persistent, a family must contain exactly 1 "
            "member in at least 99.0% of all genomes. The other genomes are absent from the "
            "family.") in out
    assert "Saving all information to a binary file for later use" in out
    assert "Generating Persistent genome of a dataset containing 4 genomes" in out
    assert ("The persistent genome contains 5 families, each one having exactly "
            "1 member from at least 99.0% of the 4 different genomes (that is 3 genomes). "
            "The other genomes are absent from the family.") in out


def test_main_pers_floor_mixed_debug(capsys):
    """
    Test that with default parameters, it creates the expected core genome.
    """
    tol = 0.99
    multi = False
    mixed = True
    floor = True
    cmd = "cmd"
    out_pers = os.path.join(GENEPATH, "PersGenome_pangenome.lst_F0.99-mixed.lst")
    assert corepers.main(cmd, UPAN, tol, multi, mixed, GENEPATH,
                         floor=floor, verbose = 15) == out_pers
    # Check creation of binary file for pangenome, and remove it
    assert os.path.isfile(UPAN + ".bin")
    # Check presence of persistent genome, and its content, and remove it
    exp_pers = os.path.join(EXP_PATH, "exp_pers-floor-mixed.txt")
    assert os.path.isfile(out_pers)
    assert tutil.compare_order_content(out_pers, exp_pers)
    # Check presence of log files and remove them
    logfile = os.path.join(GENEPATH, "PanACoTA-corepers.log")
    assert os.path.isfile(logfile)
    assert os.path.isfile(logfile + ".details")
    assert os.path.isfile(logfile + ".debug")
    assert os.path.isfile(logfile + ".err")
    # Check log messages
    out, err = capsys.readouterr()
    assert ("Will generate a Persistent genome with member(s) in at least 99.0% "
            "of all genomes in each family.") in out
    assert ("Mixed families are allowed. To be considered as persistent, "
            "a family must have exactly 1 member in 99.0% of the genomes, but in the "
            "remaining 1.0% genomes, there can be 0, 1 or several members.") in out
    assert "Saving all information to a binary file for later use" in out
    assert "Generating Persistent genome of a dataset containing 4 genomes" in out
    assert ("The persistent genome contains 7 families, each one having exactly "
            "1 member from at least 99.0% of the genomes (3 genomes). In the "
            "remaining 1.0% genomes, there can be 0, 1 or several members.") in out


def test_main_pers_floor_multi(capsys):
    """
    Test that with default parameters, it creates the expected core genome.
    """
    tol = 0.99
    multi = True
    mixed = False
    floor = True
    cmd = "cmd"
    outdir = os.path.join(GENEPATH, "outdir")
    out_pers = os.path.join(outdir, "PersGenome_pangenome.lst_F0.99-multi.lst")
    assert corepers.main(cmd, UPAN, tol, multi, mixed, outdir, floor=floor) == out_pers
    # Check creation of binary file for pangenome, and remove it
    assert os.path.isfile(UPAN + ".bin")
    # Check presence of persistent genome, and its content, and remove it
    exp_pers = os.path.join(EXP_PATH, "exp_pers-floor-multi.txt")
    assert os.path.isfile(out_pers)
    assert tutil.compare_order_content(out_pers, exp_pers)
    # Check presence of log files and remove them
    logfile = os.path.join(outdir, "PanACoTA-corepers.log")
    assert os.path.isfile(logfile)
    assert os.path.isfile(logfile + ".err")
    # Check log messages
    out, err = capsys.readouterr()
    assert ("Will generate a Persistent genome with member(s) in at least 99.0% "
            "of all genomes in each family.") in out
    assert ("Multigenic families are allowed (several members in any genome of a family).") in out
    assert "Saving all information to a binary file for later use" in out
    assert "Generating Persistent genome of a dataset containing 4 genomes" in out
    assert ("The persistent genome contains 8 families with members present in "
            "at least 3 different genomes (99.0% of the total number of genomes).") in out


def test_main_from_parse(capsys):
    """
    Test main when we give the output of the parser
    """
    import argparse
    tol = 1
    multi = False
    mixed = False
    floor = False
    args = argparse.Namespace()
    args.pangenome = UPAN
    args.tol = tol
    args.multi = multi
    args.mixed = mixed
    args.floor = floor
    args.outputdir = GENEPATH
    args.verbose = 0
    args.quiet = False
    args.argv = "PanACoTA corepers test_main_from_parse"

    corepers.main_from_parse(args)

    # Check creation of binary file for pangenome, and remove it
    out_pers = os.path.join(GENEPATH, "PersGenome_pangenome.lst_1.lst")
    assert os.path.isfile(UPAN + ".bin")
    # Check presence of persistent genome, and its content, and remove it
    exp_pers = os.path.join(EXP_PATH, "exp_coregenome.txt")
    assert os.path.isfile(out_pers)
    assert tutil.compare_order_content(out_pers, exp_pers)
    # Check presence of log files and remove them
    logfile = os.path.join(GENEPATH, "PanACoTA-corepers.log")
    assert os.path.isfile(logfile)
    assert os.path.isfile(logfile + ".err")
    # Check log messages
    out, err = capsys.readouterr()
    print(out)
    assert "Will generate a CoreGenome." in out
    assert "Saving all information to a binary file for later use" in out
    assert "Generating Persistent genome of a dataset containing 4 genomes" in out
    assert ("The core genome contains 2 families, each one having exactly 4 "
            "members, from the 4 different genomes.") in out
